"""
Relevant name changes:
- [logical] buffer/lbuf -> buffet
- [logical] comp/lcomp -> compute_einsum
-
"""

from abc import ABC

from collections import defaultdict
from dataclasses import dataclass
from typing import List, TypeAlias

import islpy as isl

from accelforge.frontend.mapping import Compute, MappingNode
from accelforge.frontend.workload import TensorName
from accelforge.model._looptree.types import Buffet

# Mapper intermediates.
##
#   @brief Iteration -> Operation relation that specifies the tiling.
#
#   The tiling relation allows us to distribute data and operations using the
#   skew and data distribution relations.
#
#   The tiling relation may have unspecified bounds which will be inferred by
#   LoopTree. The tiling relation that goes to the nest analysis is guaranteed
#   to be fully specified.
EinsumName: TypeAlias = str
"Einsum's identifier."
Tiling: TypeAlias = isl.Map
"Tiling of data and operations."
BranchTiling: TypeAlias = defaultdict[MappingNode, Tiling]
"Relation between a node and its tiling."
BuffetTiling: TypeAlias = defaultdict[Buffet, Tiling]
"Relation between a buffet and its tiling."


@dataclass(frozen=True, slots=True)
class Tag(ABC):  # pylint: disable=too-few-public-methods
    """Associating an element with its type metadata without introspection?"""


class TemporalTag(Tag):  # pylint: disable=too-few-public-methods
    """The associated element is temporally spreading?"""


@dataclass(frozen=True, slots=True)
class SpatialTag(Tag):  # pylint: disable=too-few-public-methods
    """The associated element is spatially spreading?"""

    spatial_dim: int
    "The spatial dim in a given buffer?"
    buffer: MappingNode
    "The buffer the spatial dim is across?"


class PipelineTag(Tag):  # pylint: disable=too-few-public-methods
    """The associated element is pipelined?"""


class SequentialTag(Tag):  # pylint: disable=too-few-public-methods
    """The associated element is serialized?"""


TEMPORAL_TAGS = (TemporalTag, SequentialTag)
BRANCH_TAGS = (PipelineTag, SequentialTag)
LOOP_TAGS = (TemporalTag, SpatialTag)


@dataclass(frozen=True, slots=True)
class TaggedMap:  # pylint: disable=too-few-public-methods
    """A :class:`isl.Map` with its dimensions tagged."""

    tags: List[Tag]
    map_: isl.Map

    def __repr__(self):
        return f"{type(self)}({self.tags}, {self.map_})"


class Occupancy(TaggedMap):  # pylint: disable=too-few-public-methods
    """Location of data in [logical?] hardware elements."""

    def __init__(self, tags: list[Tag], map_: isl.Map):
        assert len(tags) == map_.dim(isl.dim_type.in_), (
            "Occupancy labels input dims with tags\n"
            "-------------------------------------\n"
            f"tags: {tags}\n"
            f"map: {map_}\n"
        )
        super().__init__(tags, map_)


class OperationOccupancy(TaggedMap):  # pylint: disable=too-few-public-methods
    """Location of operations in [logical?] hardware elements."""


class Fill(TaggedMap):
    """Spacetime -> fill of a logical buffer"""

    def __init__(self, tags: list[Tag], map_: isl.Map):
        assert len(tags) == map_.dim(isl.dim_type.in_), (
            "Fill labels input dims with tags\n"
            "--------------------------------\n"
            f"tags: {tags}\n"
            f"map: {map_}\n"
        )
        super().__init__(tags, map_)


class Skew(TaggedMap):  # pylint: disable=too-few-public-methods
    """TODO: Figure out what this is."""


@dataclass(frozen=True, slots=True)
class BufferTensorEinsum:
    """
    A buffet relating a [logical?] hardware element storing data, a tensor it
    contains, and the [logical?] hardware element that is requesting the tensor.

    See Also:
    ---------
    :class:`accelforge.model._looptree.reuse.Buffet`
    """

    buffer: str
    "The logical name of the buffer supplying the tensor."
    tensor: TensorName
    "The tensor being supplied."
    einsum: Compute
    "The leaf in mapping doing the einsum compute on tensor."


@dataclass(frozen=True, slots=True)
class ComputeEinsum:
    """A logical computation the workload? needs to carry out."""

    compute: str
    """TODO: Figure out what this does."""
    branch_leaf_node: Compute
    """TODO: The compute element at the leaf of a :class:`BranchTiling`"""


# Output classes.
@dataclass(frozen=True, slots=True)
class SkewsInfo:  # pylint: disable=too-few-public-methods
    """TODO: Figure out what this does."""

    bte_to_skew: defaultdict[BufferTensorEinsum, Skew]
    """Relates a :class:`~.BufferTensorEinsum` to a :class:`~.Skew`"""
    ce_unit_to_skew: defaultdict[ComputeEinsum, Skew]
    """Relates a :class:`~.ComputeEinsum` to a :class:`~.Skew`"""


@dataclass(frozen=True, slots=True)
class MappingAnalysisResult:  # pylint: disable=too-few-public-methods
    """
    Results of mapping analysis that will become input into reuse
    analysis.
    """

    buffet_direct_above_sequential: defaultdict[Buffet, bool]
    """
    Whether a buffet is right above a sequential node. This is used when calculating
    capacity since some data can be dropped earlier than usual when using sequential
    mapping without tiling.
    """
    buffet_to_occupancy: defaultdict[BufferTensorEinsum, Occupancy]
    """The occupancy of every buffet as defined in the mapping."""
    compute_einsum_to_occupancy: defaultdict[ComputeEinsum, OperationOccupancy]
    """The occupancy of every compute unit."""
    # TODO: Figure out if this is deprecated:
    # https://github.com/NVlabs/timeloop/blob/32370826fdf1aa3c8deb0c93e6b2a2fc7cf053aa/include/loop-analysis/mapping-to-isl/fused-mapping-to-isl.hpp#L31-L35
    # node_to_buffets
    # Buffets found between the current root/branch node and the next one.
    branch_tiling: BranchTiling
    """
    Tiling of each branch. The tiling is a relation between tiling variables and
    operations. An uncompletely tiled branch will have multiple-valued :class:`isl.Map`.
    """
    compute_to_assumed_parallelism: defaultdict[MappingNode, float]
    """
    We can assume an amount of parallelism to quickly calculate approx. compute
    latency by simply dividing number of operations with assumed parallelism.
    """



from enum import auto, Flag

from functools import reduce
from operator import or_


class Metrics(Flag):
    """
    Metrics used to optimize mappings or reported by model.
    """

    LATENCY = auto()
    """The amount of time taken to execute the workload. """

    ENERGY = auto()
    """The amount of energy consumed by the workload. """

    DYNAMIC_ENERGY = auto()
    """The amount of dynamic energy consumed by the workload. """

    LEAK_ENERGY = auto()
    """The amount of leak energy consumed by the workload. """

    RESOURCE_USAGE = auto()
    """
    The amount of resources used by the workload.

    When used as a mapper objective, this objective is multivariate, and must
    consider every resource available to the hardware.
    """

    ACTIONS = auto()
    """Action counts."""

    DETAILED_MEMORY_USAGE = auto()
    """
    Memory usage broken down by tensor and Einsum.
    """

    @classmethod
    def all_metrics(cls):
        return reduce(or_, iter(cls), cls.LATENCY)

    def includes_leak_energy(self) -> bool:
        """Returns True if the metrics include leak energy, either alone or as part of
        total energy. False otherwise."""
        return self & (Metrics.ENERGY | Metrics.LEAK_ENERGY)

    def includes_dynamic_energy(self) -> bool:
        """Returns True if the metrics include dynamic energy, either alone or as part
        of total energy. False otherwise."""
        return self & (Metrics.ENERGY | Metrics.DYNAMIC_ENERGY)



def compute_max(child_caps, caps):
    buf_to_max = {}
    for child_cap in child_caps:
        buf_to_child_usage = {}

        for buf, v in child_cap.items():
            if buf not in buf_to_child_usage:
                buf_to_child_usage[buf] = 0
            if buf not in buf_to_max:
                buf_to_max[buf] = 0

            buf_to_child_usage[buf] += v

        for buf in buf_to_child_usage:
            buf_to_max[buf] = max(buf_to_max[buf], buf_to_child_usage[buf])

    for buf, c in buf_to_max.items():
        if buf not in caps:
            caps[buf] = 0
        caps[buf] += c


def compute_total(child_caps, caps):
    for child_cap in child_caps:
        for buf, v in child_cap.items():
            if buf in caps:
                caps[buf] += v
            else:
                caps[buf] = v


CAPACITY_AGGREGATORS = {
    "sequential": compute_max,
    "pipeline": compute_total,
    "parallel": compute_total,
}



from collections import defaultdict
import copy
import functools
import itertools

from typing import Any, Callable, Iterable

import sympy

from accelforge.frontend.mapping import Nested, TilePattern
from accelforge.frontend.mapping import Loop as MappingLoop
from accelforge.mapper.FFM._join_pmappings.compatibility import (
    Compatibility,
    Loop,
    TensorReservation,
)
from accelforge.util import _fillna_and__numeric_cast, _numeric_cast
from accelforge.util._frozenset import fzs, oset

from accelforge._accelerated_imports import pd

from accelforge.mapper.FFM._pareto_df.df_convention import *
from accelforge.mapper.FFM._pareto_df.pareto import makepareto

CHECK_CORRECTNESS = False
DEBUG_PRINT_NO_VALID = False


def error_check_wrapper(func):
    if not CHECK_CORRECTNESS:
        return func

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            prev_args, prev_kwargs = copy.deepcopy(args), copy.deepcopy(kwargs)
            return func(*args, **kwargs)
        except Exception as e:
            print(f"EXCEPTION: {e}")
            live_tensors = oset()
            if "live_tensors" in kwargs:
                live_tensors = kwargs["live_tensors"]
            else:
                argnames = func.__code__.co_varnames[: func.__code__.co_argcount]
                if "live_tensors" in argnames:
                    idx = argnames.index("live_tensors")
                    if idx < len(args):
                        live_tensors = args[idx]
            for prev_arg in itertools.chain(prev_args, prev_kwargs.values()):
                if isinstance(prev_arg, PmappingDataframe):
                    prev_arg.fail(0, live_tensors)
                break
            func(*args, **kwargs)  # For debugging

    return wrapper


def reduce_precision(data: pd.DataFrame) -> pd.DataFrame:
    data = _numeric_cast(data)

    # def _reduce_precision(c: str, s: pd.Series) -> pd.Series:
    #     # If it's an int type, check the range. If within range of 8b change to 8b. If
    #     # within the range of 16b change to 16b...

    #     # If it's a float, cast to NUMPY_FLOAT_TYPE
    #     if pd.api.types.is_float_dtype(s) and s.dtype != NUMPY_FLOAT_TYPE:
    #         return s.astype(NUMPY_FLOAT_TYPE)

    #     if not is_fused_loop_col(c):
    #         return s

    #     # Get the range of the column
    #     min_val = s.min()
    #     if min_val < 0:
    #         return s

    #     max_val = s.max()
    #     if max_val <= 2**8 - 1 and s.dtype != np.uint8:
    #         return s.astype(np.uint8)
    #     elif max_val <= 2**16 - 1 and s.dtype != np.uint16:
    #         return s.astype(np.uint16)
    #     elif max_val <= 2**32 - 1 and s.dtype != np.uint32:
    #         return s.astype(np.uint32)
    #     return s

    # for c in data.columns:
    #     data.loc[:, c] = _reduce_precision(c, data.loc[:, c])

    return data


def get_reservation_or_parent(
    name: str,
    level: int,
    l_reservations: dict[str, set[int]],
    r_reservations: dict[str, set[int]],
    left: bool = False,
    return_name_level_left: bool = False,
) -> str | tuple[str, int, bool] | None:
    reservations = l_reservations if left else r_reservations
    if (reservations := reservations.get(name, None)) is not None:
        while level >= -1:
            if level in reservations:
                if return_name_level_left:
                    return name, level, left
                return reservation2col(name, level, left)
            # The parent of left nodes are right nodes, so if we don't find a
            # left node immediately then we're back on the right nodes
            reservations = r_reservations.get(name, oset())
            left = False
            level -= 1
    return None


class PmappingDataframe:
    def __init__(
        self,
        data: pd.DataFrame,
        n_total_pmappings: float,
        n_valid_pmappings: float,
        ignored_resources: set[str],
        drop_valid_reservations: bool,
        skip_pareto: bool = False,
        fill_reservation_cols: set | str = fzs(),
        check_above_subset_below: bool = CHECK_CORRECTNESS,
        max_right_to_left: bool = False,
        next_shared_loop_index: int = None,
        excess_resource_tolerance: float = 0,
    ):
        self._data: pd.DataFrame = reduce_precision(data)
        self._prev_free_to_loop_index = None
        self.n_total_pmappings: float = n_total_pmappings
        self.n_valid_pmappings: float = n_valid_pmappings
        self.drop_valid_reservations: bool = drop_valid_reservations
        self.excess_resource_tolerance: float = excess_resource_tolerance

        if next_shared_loop_index is not None:
            assert (
                ignored_resources is not None
            ), "ignored_resources must be set if next_shared_loop_index is set"
            self.free_to_loop_index(next_shared_loop_index)
            self.limit_capacity(
                next_shared_loop_index=next_shared_loop_index,
                ignored_resources=ignored_resources,
            )

        if fill_reservation_cols:  # Affects PmappingDataframe so must go before
            self.fill_reservation_cols(fill_reservation_cols)
        if check_above_subset_below:
            self.check_above_subset_below()
        if max_right_to_left:  # Affects PmappingDataframe so must go before
            self.max_right_to_left()
        if check_above_subset_below:
            self.check_above_subset_below()

        if not skip_pareto:
            self.make_pareto()

        if check_above_subset_below:
            self.check_above_subset_below()

        self.ignored_resources = ignored_resources

        assert len(self.data.columns) == len(
            oset(self.data.columns)
        ), f"Duplicate columns: {self.data.columns}"

    def rename(self, renames: dict[str, str]) -> "PmappingDataframe":
        new = self.copy()
        new.data.rename(columns=renames, inplace=True)
        return new

    @error_check_wrapper
    def fill_reservation_cols(self, columns: set | str):
        l_reservations, r_reservations = self._make_reservations()
        targets = []
        if columns == "auto":
            for left, reservations_dict in [
                (True, l_reservations),
                (False, r_reservations),
            ]:
                for resource, reservations in reservations_dict.items():
                    for r in sorted(reservations):
                        above = get_reservation_or_parent(
                            resource, r - 1, l_reservations, r_reservations
                        )
                        if above is not None:
                            below = reservation2col(resource, r, left=left)
                            targets.append((r, above, below))
        else:
            for below in columns:
                if (name_nloops := col2reservation(below)) is None:
                    raise ValueError(f"{below} is not a valid reservation column")
                name, nloops = name_nloops.name, name_nloops.nloops
                above = get_reservation_or_parent(
                    name, nloops - 1, l_reservations, r_reservations
                )
                if above is not None:
                    targets.append((nloops, above, below))

        # Sort so we go from top to bottom. Needed in case we have to max 0->1
        # then 1->2
        for _, above, below in sorted(targets, key=lambda x: x[0]):
            assert (
                above in self.data.columns
            ), f"Missing column {above}. Have columns:\n\t" + "\n\t".join(
                list(self.data.columns)
            )
            assert (
                below in self.data.columns
            ), f"Missing column {below}. Have columns:\n\t" + "\n\t".join(
                list(self.data.columns)
            )
            max_to_col(self.data, below, above)

    @error_check_wrapper
    def max_right_to_left(self):
        l_reservations, r_reservations = self._make_reservations()
        for resource, reservations in l_reservations.items():
            for r in reservations:
                if r in r_reservations.get(resource, oset()):
                    source = reservation2col(resource, r)
                    target = reservation2col(resource, r, left=True)
                    max_to_col(self.data, target, source)

    @property
    def data(self) -> pd.DataFrame:
        return self._data

    @error_check_wrapper
    def _make_reservations(self) -> tuple[dict[str, set[int]], dict[str, set[int]]]:
        """
        Create a dictionary of reservations for each resource.
        The dictionary keys are the resource names and the values are lists
        of column names for each loop index.
        """
        l_reservations, r_reservations = {}, {}
        for c in self.data.columns:
            if (key := col2reservation(c)) is not None:
                target = l_reservations if is_left_col(c) else r_reservations
                target.setdefault(key.name, oset()).add(key.nloops)
                assert key.nloops >= -1

        return l_reservations, r_reservations

    def clear_fused_loop_symbols(self):
        dropcols = [c for c in self.data.columns if is_fused_loop_col(c)]
        if not dropcols:
            return
        self.data.drop(columns=dropcols, inplace=True)
        self.make_pareto()

    @error_check_wrapper
    def free_to_loop_index(self, loop_index: int) -> bool:
        """
           A  B
            / | --- 0
           C  D
            / | --- 1  < Shared Loop Index
           E  F
            / | --- 2
           G  H
        ->
           A  B
            / | --- 0
           C  D
              | --- 1  < Shared Loop Index
          max(E,G,H)
        We skip incorporating E into the max because its reservations are
        already incorporated into F and G.
        """
        if loop_index == self._prev_free_to_loop_index:
            return False
        self._prev_free_to_loop_index = loop_index

        drop_columns = []
        l_reservations, r_reservations = self._make_reservations()
        for resource in oset(l_reservations) | oset(r_reservations):
            max_columns = []
            cur_l_reservations = l_reservations.get(resource, oset())
            cur_r_reservations = r_reservations.get(resource, oset())
            left_big_enough = [l for l in cur_l_reservations if l >= loop_index + 1]
            right_big_enough = [
                r for r in cur_r_reservations if r >= loop_index + 2
            ]  # + 1 is target

            if len(right_big_enough) > 1:  # All ones above the last are subsets
                right_biggest = max(right_big_enough)
                right_big_enough.remove(right_biggest)
                drop_columns += [reservation2col(resource, r) for r in right_big_enough]
                right_big_enough = [right_biggest]

            max_columns = [reservation2col(resource, r) for r in right_big_enough] + [
                reservation2col(resource, l, left=True) for l in left_big_enough
            ]

            if not max_columns:
                continue

            target = reservation2col(resource, loop_index + 1)
            if target in self.data:
                max_columns.append(target)

            if len(max_columns) == 1:
                self.data.rename(columns={max_columns[0]: target}, inplace=True)
            else:
                for c in max_columns:
                    max_to_col(self.data, target, c)
                drop_columns += [m for m in max_columns if m != target]
        self.data.drop(columns=drop_columns, inplace=True)

        return len(drop_columns) != 0

    @error_check_wrapper
    def shift_bottom_reservation_left(self, bottom_loop_index: int):
        """
        Shifts the bottom reservation from right to left.
        Example:
            Before:                After:
            A  B                   A  B
             / | --- 0             / | --- 0
            C  D                  C  D
               | --- 1             /   --- 1
               E                  E
        """
        _l_reservations, r_reservations = self._make_reservations()
        for resource in r_reservations:
            if bottom_loop_index + 1 not in r_reservations[resource]:
                continue
            source = reservation2col(resource, bottom_loop_index + 1)
            target = reservation2col(resource, bottom_loop_index + 1, left=True)
            if target in self.data:
                max_to_col(self.data, target, source)
                self.data.drop(columns=[source], inplace=True)
            else:
                self.data.rename(columns={source: target}, inplace=True)

    @staticmethod
    def _get_target_path(suffix: str = None) -> str:
        import os

        f = "./images"
        os.makedirs(f, exist_ok=True)
        suffix = "" if suffix is None else f".{suffix}"
        i = 0
        while os.path.exists(os.path.join(f, f"test_{i}{suffix}.png")):
            i += 1
        return os.path.join(f, f"test_{i}{suffix}.png")

    def get_max_loop_index(self):
        l_reservations, r_reservations = self._make_reservations()
        return max(
            max(
                (max(r, default=-1) for r in r_reservations.values()),
                default=-1,
            ),
            max(
                (max(r, default=-1) for r in l_reservations.values()),
                default=-1,
            ),
        )

    def get_min_loop_index(self):
        l_reservations, r_reservations = self._make_reservations()
        return min(
            min(
                (min(r, default=1000000) for r in r_reservations.values()),
                default=1000000,
            ),
            min(
                (min(r, default=1000000) for r in l_reservations.values()),
                default=1000000,
            ),
        )

    @error_check_wrapper
    def merge_next(
        self,
        right: "PmappingDataframe",
        duplicated_aliased_tensors: set[TensorReservation],
        compatibility_left: Compatibility,
        compatibility_right: Compatibility,
        compatibility_joined: Compatibility,
        ignored_resources: set[str],
        _pmapping_row_filter_function: Callable[[pd.Series], bool] | None = None,
        _force_allow_invalid_only_for_runtime_test: bool = False,
        _is_invalid: bool = False,
    ) -> "PmappingDataframe":
        """
           A  B            A2
            / | --- 0      |
           C  D            C2
              | --- 1      |     < Shared Loop Index
              E            E2
                           |
                           F2
           ->
           A  A+A2
            / | --- 0
        C+A2  C+C2
            / | --- 1  < Shared Loop Index
        E+C2  E2+D
              |
              F2+D
        """
        live_tensors = compatibility_joined.tensor_names
        shared_loop_index = compatibility_left.n_loops - 1
        next_shared_loop_index = compatibility_joined.n_loops - 1

        self.free_to_loop_index(shared_loop_index)
        self.shift_bottom_reservation_left(shared_loop_index)

        shared_tensor_names = (
            compatibility_left.tensor_names & compatibility_right.tensor_names
        )
        shared_tensors = [
            compatibility_left.get_reservation_of_tensor(s) for s in shared_tensor_names
        ]
        left_match, right_match = [], []
        make_empty_result = False

        def check_match(la: Loop, lb: Loop, param: str):
            a, b = getattr(la.tile_pattern, param), getattr(lb.tile_pattern, param)
            if isinstance(a, str) or isinstance(b, str):
                left_match.append(a)
                right_match.append(b)
            elif a != b:
                raise ValueError(f"Mismatch in {param}: {a} != {b}")

        try:
            for s in shared_tensor_names:
                ta = compatibility_left.get_reservation_of_tensor(s)
                tb = compatibility_right.get_reservation_of_tensor(s)
                for la, lb in zip(ta.loops, tb.loops):
                    check_match(la, lb, "initial_tile_shape")
                    check_match(la, lb, "tile_shape")

            for la, lb in zip(compatibility_left.loops, compatibility_right.loops):
                check_match(la, lb, "calculated_n_iterations")

        except ValueError as e:
            make_empty_result = True

        right_df_l_reservations, right_df_r_reservations = right._make_reservations()
        assert not right_df_l_reservations, f"{right_df_l_reservations} is not None"

        l_reservations, r_reservations = self._make_reservations()

        for resource, reservations in r_reservations.items():
            n_reservations = max(reservations, default=-1)
            assert (
                n_reservations <= shared_loop_index
            ), f"{resource}: {reservations} > {shared_loop_index}"

        for resource, reservations in l_reservations.items():
            n_reservations = max(reservations, default=-1)
            assert (
                n_reservations <= shared_loop_index + 1
            ), f"{resource}: {reservations} > {shared_loop_index}"

        max_nloops = max(
            shared_loop_index, self.get_max_loop_index(), right.get_max_loop_index()
        )
        min_nloops = min(self.get_min_loop_index(), right.get_min_loop_index())

        sd, rd = self.data, right.data
        if make_empty_result:
            sd = sd.iloc[0:0]
            rd = rd.iloc[0:0]

        # _force_allow_invalid_only_for_runtime_test -> only merge matched ones
        if left_match and not _force_allow_invalid_only_for_runtime_test:
            df = pd.merge(
                sd,
                rd,
                how="inner",
                left_on=left_match,
                right_on=right_match,
                suffixes=["", "_RIGHT_MERGE"],
            )
        else:
            df = pd.merge(sd, rd, how="cross", suffixes=["", "_RIGHT_MERGE"])

        df = reduce_precision(df)

        # Drop all fused loop columns that are not used anymore
        remaining_symbols = compatibility_joined.symbols()
        dropcols = [
            c for c in df.columns if is_fused_loop_col(c) and c not in remaining_symbols
        ]
        df = df.drop(columns=dropcols)

        # Number of combinations
        n_total_pmappings = self.n_total_pmappings * right.n_total_pmappings
        n_valid_pmappings = self.n_valid_pmappings * right.n_valid_pmappings
        scale_by = len(df) / max(1, len(self.data) * len(right.data))
        n_total_pmappings *= scale_by
        n_valid_pmappings *= scale_by

        # Make sure everything is done in increasing loop order so we don't have
        # read-after-write hazards
        for nloops in range(max_nloops, min_nloops - 1, -1):

            def iter_reservations(reservations_dict):
                for resource in reservations_dict:
                    if nloops in reservations_dict[resource]:
                        yield resource

            # For the RIGHT tree, RIGHT reservations: If there is no matching node in
            # the left tree, add the above-this-level reservation from the left tree. If
            # there is a matching node in the left tree, then we'll add this node to it
            # in the next step.
            for resource in iter_reservations(right_df_r_reservations):
                if (
                    source := get_reservation_or_parent(
                        resource, nloops - 1, l_reservations, r_reservations
                    )
                ) is None:
                    continue
                target = reservation2col(resource, nloops)
                # If there's a merged version column, then it's in both trees
                if target + "_RIGHT_MERGE" in df:
                    continue
                add_to_col(df, target, source)
            # For LEFT tree, LEFT reservations: Add the immediately-above
            # reservation from the right tree.
            for resource in iter_reservations(l_reservations):
                if (
                    source := get_reservation_or_parent(
                        resource,
                        nloops - 1,
                        right_df_l_reservations,
                        right_df_r_reservations,
                    )
                ) is None:
                    continue
                right_merge_source = source + "_RIGHT_MERGE"
                target = reservation2col(resource, nloops, left=True)
                if source is not None:
                    add_to_col(
                        df,
                        target,
                        right_merge_source if right_merge_source in df else source,
                    )
            # For LEFT tree, RIGHT reservations: Add the same-level reservation from the
            # right tree. This will double-count reservations that are in both branches,
            # so we remove them later.
            for resource in iter_reservations(r_reservations):
                if (
                    source := get_reservation_or_parent(
                        resource,
                        nloops,
                        right_df_l_reservations,
                        right_df_r_reservations,
                    )
                ) is None:
                    continue
                right_merge_source = source + "_RIGHT_MERGE"
                target = reservation2col(resource, nloops)
                if source is not None:
                    add_to_col(
                        df,
                        target,
                        right_merge_source if right_merge_source in df else source,
                    )

        # For everything else: Simple add
        dropcols = [c for c in df.columns if c.endswith("_RIGHT_MERGE")]
        for source in dropcols:
            target = source[: -len("_RIGHT_MERGE")]
            if is_tensor_col(target):
                continue
            if not col_used_in_pareto(target):
                raise ValueError(f"{target} is not used in pareto")
            if col2reservation(target) is None:
                add_to_col(df, target, source)

        df = df.drop(columns=dropcols)
        result = PmappingDataframe(
            df,
            skip_pareto=True,
            check_above_subset_below=False,
            n_total_pmappings=n_total_pmappings,
            n_valid_pmappings=n_valid_pmappings,
            ignored_resources=self.ignored_resources,
            drop_valid_reservations=self.drop_valid_reservations,
        )
        # Remove tensors that were allocated in both branches and got added
        # together.
        shared_to_free = [
            s for s in shared_tensors if s.above_loop_index <= shared_loop_index
        ]
        reservations_of_live_tensor_not_in_right = [
            compatibility_joined.get_reservation_of_tensor(t)
            for t in compatibility_joined.tensor_names
            - compatibility_right.tensor_names
        ]
        live_to_alloc = [
            r
            for r in reservations_of_live_tensor_not_in_right
            if r.above_loop_index > shared_loop_index
        ]
        result.adjust_reservations(
            alloc=live_to_alloc,
            free=list(itertools.chain(shared_to_free, duplicated_aliased_tensors)),
            ignored_resources=ignored_resources,
        )

        if CHECK_CORRECTNESS:
            result.check_above_subset_below(live_tensors)
            result.check_reservations(live_tensors)

        result.free_to_loop_index(next_shared_loop_index)
        if not CHECK_CORRECTNESS:
            result.limit_capacity(
                next_shared_loop_index, ignored_resources=ignored_resources
            )
        result.max_right_to_left()
        if _pmapping_row_filter_function is not None:
            result = result.filter_rows(_pmapping_row_filter_function)
        result.make_pareto()

        # This join was invalid (we only ran it for runtime measurement). Clear all the
        # mappings.
        if _is_invalid:
            result._data = result.data.iloc[0:0]

        # If we're running _force_allow_invalid_only_for_runtime_test, we don't want to
        # have all those invalid combinations, so return the actual merge result.
        elif _force_allow_invalid_only_for_runtime_test:
            return self.merge_next(
                right,
                duplicated_aliased_tensors,
                compatibility_left,
                compatibility_right,
                compatibility_joined,
                ignored_resources,
                _pmapping_row_filter_function,
                _force_allow_invalid_only_for_runtime_test=False,
            )

        return result

    @error_check_wrapper
    def _adjust_reservations_one_resource(
        self,
        resource: str,
        alloc: Iterable[TensorReservation],
        free: Iterable[TensorReservation],
    ):
        alloc, free = list(alloc), list(free)
        # Iterate through each reservation and level
        targets = defaultdict(int)

        l_reservations, r_reservations = self._make_reservations()

        # Must allocate at the above_loop_index level
        for t in itertools.chain(alloc, free):
            r_reservations.setdefault(resource, oset()).add(t.above_loop_index)

        for t, negate in [(t, False) for t in alloc] + [(t, True) for t in free]:
            size = self.data[tensor2col(t.name)]
            size = -size if negate else size
            targets[t.above_loop_index, False] += size
            # Allocate at any levels below the above_loop_index level
            for level in r_reservations[resource]:
                if level > t.above_loop_index:
                    targets[level, False] += size
            for level in l_reservations.get(resource, oset()):
                if level > t.above_loop_index:
                    targets[level, True] += size

        # Now apply the allocations. Sort so we go from top to bottom in case
        # there are maxes that propagate down.
        for (level, left), size in sorted(
            targets.items(), key=lambda x: x[0], reverse=True
        ):
            target = reservation2col(resource, level, left=left)
            if target in self.data:
                add_to_col(self.data, target, size)
                continue

            # We're creating a new column, so copy allocations from any parents
            source = get_reservation_or_parent(
                resource, level - 1, l_reservations, r_reservations
            )
            if source:
                assert (
                    source in self.data.columns
                ), f"{source} not in {sorted(self.data.columns)}"
                add_to_col(self.data, target, source)
                add_to_col(self.data, target, size)
            else:
                self.data[target] = size

            # We made a new column! Update our reservations so future iterations
            # know about it.
            l_reservations, r_reservations = self._make_reservations()

            # Assert all reservations are >= 0
            assert (self.data[target] >= 0).all(), f"Negative reservation: {target}"

    @error_check_wrapper
    def adjust_reservations(
        self,
        alloc: Iterable[TensorReservation],
        free: Iterable[TensorReservation],
        ignored_resources: set[str] = oset(),
    ):
        alloc, free = list(alloc), list(free)
        all_resources = oset(t.resource_name for t in alloc) | oset(
            t.resource_name for t in free
        )
        ignored_resources = ignored_resources | self.ignored_resources
        # Handle each resource separately
        for resource in all_resources:
            if resource in ignored_resources:
                continue
            cur_alloc = [t for t in alloc if t.resource_name == resource]
            cur_free = [t for t in free if t.resource_name == resource]
            if cur_alloc or cur_free:
                self._adjust_reservations_one_resource(resource, cur_alloc, cur_free)

    @staticmethod
    def concat(
        paretos: list["PmappingDataframe"], skip_pareto: bool = False
    ) -> "PmappingDataframe":
        if len(paretos) == 0:
            raise ValueError("No paretos to concatenate")
        if len(paretos) == 1:
            return paretos[0]

        required_cols = oset.union(*[oset(p.data.columns) for p in paretos])
        shared_cols = oset.intersection(*[oset(p.data.columns) for p in paretos])
        fill_cols = required_cols - shared_cols
        fill_cols = [c for c in fill_cols if col_used_in_pareto(c)]

        concatenated = pd.concat([p.data for p in paretos]).reset_index(drop=True)

        p = PmappingDataframe(
            _fillna_and__numeric_cast(concatenated, 0),
            skip_pareto=len(paretos) == 1 or skip_pareto,
            fill_reservation_cols=fill_cols,
            n_total_pmappings=sum(p.n_total_pmappings for p in paretos),
            n_valid_pmappings=sum(p.n_valid_pmappings for p in paretos),
            ignored_resources=next(iter(paretos)).ignored_resources,
            drop_valid_reservations=next(iter(paretos)).drop_valid_reservations,
        )
        return p

    def update(
        self,
        skip_pareto: bool,
        **kwargs,
    ) -> "PmappingDataframe":
        args = dict(
            data=self.data,
            skip_pareto=skip_pareto,
            check_above_subset_below=False,
            n_total_pmappings=self.n_total_pmappings,
            n_valid_pmappings=self.n_valid_pmappings,
            ignored_resources=self.ignored_resources,
            drop_valid_reservations=self.drop_valid_reservations,
        )
        args.update(kwargs)
        return PmappingDataframe(**args)

    def copy(self, copy_df: bool = True) -> "PmappingDataframe":
        return self.update(
            data=self.data.copy() if copy_df else self.data,
            skip_pareto=True,
            check_above_subset_below=False,
        )

    def limit_capacity(
        self,
        next_shared_loop_index: int = None,
        ignored_resources: set[str] = oset(),
        finished: bool = False,
    ):
        dropcols = []
        l_reservations, r_reservations = self._make_reservations()
        tolerance = self.excess_resource_tolerance
        for resource in sorted(oset(r_reservations) | oset(l_reservations)):
            # Right reservations: Only check the greatest-index level. If a loop
            # is 0 and the next shared loop index is -1, then we can drop the
            # column.
            right_loops = r_reservations.get(resource, oset())
            for l in list(right_loops):
                col = reservation2col(resource, l)
                if (
                    DEBUG_PRINT_NO_VALID
                    and sum(self.data[col] <= 1 + tolerance) == 0
                    and len(self.data) == 1
                    and tolerance == 0
                ):
                    print(
                        f"Resource {resource} has no valid reservations. Failed for {col}: {next(iter(self.data[col]))} <= {1 + tolerance}: {next(iter(self.data[col])) <= 1 + tolerance}"
                    )
                    for col2 in self.data.columns:
                        print(f"{col2}: {list[Any](self.data[col2])}")
                self._data = self.data[self.data[col] <= 1 + tolerance]
                if (
                    l == 0
                    and next_shared_loop_index == -1
                    # CAN'T DROP RESERVATIONS UNTIL WE'RE FINISHED JOINING. Persistent
                    # tensors may get saved later and would live at the same time as
                    # these reservations.
                    and finished
                    and self.drop_valid_reservations
                    and resource not in ignored_resources
                    and (tolerance == 0 or not any(self.data[col] > 1))
                ):
                    right_loops.discard(l)
                    dropcols.append(col)

            # Left reservations: Check all levels. If a loop is 0,
            # then we can drop the column.
            left_loops = l_reservations.get(resource, oset())
            for l in list(left_loops):
                col = reservation2col(resource, l, left=True)
                if (
                    DEBUG_PRINT_NO_VALID
                    and sum(self.data[col] <= 1 + tolerance) == 0
                    and len(self.data) == 1
                    and tolerance == 0
                ):
                    print(
                        f"Resource {resource} has no valid reservations. Failed for {col}: {next(iter(self.data[col]))} <= {1 + tolerance}: {next(iter(self.data[col])) <= 1 + tolerance}"
                    )
                    for col2 in self.data.columns:
                        print(f"{col2}: {list[Any](self.data[col2])}")
                self._data = self.data[self.data[col] <= 1 + tolerance]
                if (
                    l == 0
                    # CAN'T DROP RESERVATIONS UNTIL WE'RE FINISHED JOINING. Persistent
                    # tensors may get saved later.
                    and finished
                    and self.drop_valid_reservations
                    and resource not in ignored_resources
                    and (tolerance == 0 or not any(self.data[col] > 1))
                ):
                    left_loops.discard(l)
                    dropcols.append(col)

        self._data = self.data.drop(columns=dropcols)

    def make_pareto(
        self,
        columns: list[str] = None,
        objective_tolerance: float = 0,
        resource_usage_tolerance: float = 0,
        absolute_resource_usage_tolerance: float = 0,
        inplace: bool = True,
    ) -> "PmappingDataframe":
        # The error for absolute_resource_usage_tolerance sums each time we modify the
        # df and prune, so if we use it more, we need to use a lower threshold. The
        # max_n_einsums value assumes that absolute_resource_usage_tolerance is only
        # used for joining.
        if self.drop_valid_reservations:
            resource_usage_tolerance = objective_tolerance

        new_data = makepareto(
            self.data,
            columns,
            resource_usage_tolerance=resource_usage_tolerance,
            absolute_resource_usage_tolerance=absolute_resource_usage_tolerance,
            objective_tolerance=objective_tolerance,
        )
        if inplace:
            self._data = new_data
            return self
        else:
            return self.update(data=new_data, skip_pareto=True)

    def has_reservations(self):
        return any(col2reservation(c) is not None for c in self.data.columns)

    # ============================================================================
    # Checking functions
    # ============================================================================
    def check_above_subset_below(self, live_tensors: set[str] = fzs()):
        assert not self.data.isnull().values.any(), f"NaN in {self.data}"
        targets = []
        l_reservations, r_reservations = self._make_reservations()
        for left, reservations_dict in [
            (True, l_reservations),
            (False, r_reservations),
        ]:
            for resource, reservations in reservations_dict.items():
                for r in reservations:
                    above = get_reservation_or_parent(
                        resource, r - 1, l_reservations, r_reservations
                    )
                    if above is not None:
                        below = reservation2col(resource, r, left=left)
                        targets.append((above, below))

        for above, below in targets:
            if (self.data[below] < self.data[above]).any():
                first_failing_index = (self.data[below] < self.data[above]).idxmax()
                fail_row = self.data.iloc[first_failing_index]
                error = f"""
                {below} column is less than {above} column. A reservation at
                a level should include all reservations above it. There were {len(fail_row)} rows
                with this error. One example: {fail_row}
                """
                self.fail(first_failing_index, live_tensors)
                raise ValueError(error)

    def filter_rows(
        self, _pmapping_row_filter_function: Callable[[pd.Series], bool] | None = None
    ) -> "PmappingDataframe":
        if _pmapping_row_filter_function is None:
            return self.copy()

        # s = _pmapping_row_filter_function(self._data)
        # if s.sum() > 0:
        #     print(f"Filter rate: {s.sum() / len(s):.2%}")
        return self.update(
            data=self._data[_pmapping_row_filter_function(self._data)].copy(),
            skip_pareto=True,
        )

    def __len__(self) -> int:
        return len(self._data)

    # @error_check_wrapper
    # def check_reservations(self, live_tensors: set[int]):
    #     from accelforge.visualization.reservationtree import mappings2reservationtree
    #     assert not self.data.isnull().values.any(), f"NaN in {self.data}"

    #     self = self.copy()

    #     self.free_to_loop_index(-1)
    #     self.shift_bottom_reservation_left(-1)

    #     for i, r in self.data.iterrows():
    #         looptree = mappings2reservationtree(
    #             r[MAPPING_COLUMN],
    #             r.get(STATS, None),
    #             still_live_tensors=live_tensors
    #         )
    #         reservations = dict(looptree.get_reservations())

    #         # If r doesn't have any columns, continue. It's a copy Einsum so it has no
    #         # stats.
    #         if r.empty:
    #             continue

    #         for k, v in reservations.items():
    #             col = get_reservation_or_parent(k, 0, left=True)
    #             if str(k) == "0":
    #                 continue
    #             if col not in self.data.columns:
    #                 got = r[[c for c in self.data.columns if col2reservation(c) is not None]]
    #                 self.fail(i, live_tensors)
    #                 raise ValueError(f"Missing {k}: Expected {reservations}. Got: {got}")
    #             if r[col] != v:
    #                 got = r[[c for c in self.data.columns if col2reservation(c) is not None]]
    #                 self.fail(i, live_tensors)
    #                 looptree = mappings2reservationtree(
    #                     r[MAPPING_COLUMN],
    #                     r.get(STATS, None),
    #                     # skip_backing_tensors_in_right_branch=live_tensors,
    #                     still_live_tensors=live_tensors,
    #                 )
    #                 raise ValueError(
    #                     f"Mismatched {k}: {v} != {r[col]}. Expected {reservations}. Got: {got}"
    #                 )

    # def fail(self, index, live_tensors):
    #     from accelforge.mapper.FFM._join_pmappings.pmapping_group import TensorReservation
    #     r = self.data.iloc[index]
    #     assert not self.data.isnull().values.any(), f"NaN in {self.data}"
    #     self = self.copy()
    #     self._draw_index(index, live_tensors, self._get_target_path(suffix="fail"))
    #     all_tensors = oset(t for tn in r[MAPPING_COLUMN].values() for t in tn.tensors)
    #     all_tensors = TensorReservation.get_backing_tensors(all_tensors)
    #     for t in sorted(all_tensors):
    #         print(f"{t.__repr__()},")

    # def _draw_index(self, index: int, live_tensors, to_file: str = "test.png"):
    #     from accelforge.visualization.reservationtree import mappings2reservationtree
    #     import pydot
    #     looptree = mappings2reservationtree(
    #         self.data.iloc[index][MAPPING_COLUMN],
    #         self.data.iloc[index].get(STATS, None),
    #         still_live_tensors=live_tensors,
    #     )
    #     graph = pydot.Dot(graph_type="digraph", ranksep="0.2", nodesep="0.2")
    #     looptree.to_pydot(graph)
    #     row = self.data.iloc[index]
    #     all_data = sorted(f"{k}: {v}" for k, v in row.items() if k not in DICT_COLUMNS)
    #     data_str = "\n".join(all_data)
    #     graph.add_node(pydot.Node("data", label=data_str, shape="plaintext"))
    #     with open(to_file, "wb") as f:
    #         f.write(graph.create_png())

    def clear_irrelevant_columns(
        self, compatibility: Compatibility
    ) -> "PmappingDataframe":
        return self.update(
            data=compatibility.clear_unrelated_columns(self._data),
            skip_pareto=True,
        )


def row2pmappings(
    row: pd.Series,
    einsum_names: list[str],
    rank_variable_bounds: dict[str, dict[str, int]],
) -> list[Nested]:
    pmappings: list[Nested] = []
    for einsum_name in einsum_names:
        pmapping: Nested = copy.deepcopy(row[f"{einsum_name}<SEP>{MAPPING_COLUMN}"])
        for node in pmapping.nodes:

            def acc(s: str | None | int):
                s = s.name if isinstance(s, sympy.Symbol) else s
                return row[f"{einsum_name}<SEP>{s}"] if isinstance(s, str) else s

            if isinstance(node, MappingLoop):
                tp: TilePattern = node.tile_pattern
                node.tile_pattern = tp.update(
                    initial_tile_shape=acc(tp.initial_tile_shape),
                    tile_shape=acc(tp.tile_shape),
                )
        pmappings.append(pmapping)
        pmapping._beautify_loops(rank_variable_bounds)
    return pmappings



from dataclasses import dataclass
import numpy as np

from accelforge.util._frozenset import oset
from accelforge.frontend.mapping import (
    Mapping,
    Compute,
    Temporal,
    Spatial,
    Split,
    Nested,
)
from accelforge.frontend.workload import RankVariable, Rank, Workload


class TemporalTrace:
    rank_variable: RankVariable


def trace_iterations(mapping: Mapping, spec):
    return _trace_iterations(mapping, spec.workload)


def _trace_iterations(node, workload: Workload):
    if isinstance(node, Nested):
        if isinstance(node.nodes[-1], (Nested, Split)):
            last_result = _trace_iterations(node.nodes[-1], workload)
        elif isinstance(node.nodes[-1], Compute):
            last_result = _trace_node(node.nodes[-1], workload, None)

        for child_node in reversed(node.nodes[:-1]):
            last_result = _trace_node(child_node, workload, last_result)
        return last_result
    elif isinstance(node, Split):
        last_result = {}
        for child_node in node.nodes:
            if isinstance(child_node, Compute):
                child_result = _trace_node(child_node, workload, None)
            elif isinstance(child_node, (Nested, Split)):
                child_result = _trace_iterations(child_node, workload)
            else:
                raise NotImplementedError()

            in_child = oset(child_result.keys())
            in_last = oset(last_result.keys())
            child_shape = None
            last_shape = None
            for rank_var in in_child & in_last:
                child_trace = child_result[rank_var]
                child_shape = child_trace.shape
                last_trace = last_result[rank_var]
                last_shape = last_trace.shape
                if rank_var in last_result:
                    last_result[rank_var] = np.concatenate([last_trace, child_trace])
                else:
                    last_result[rank_var] = child_trace
            for rank_var in in_last - in_child:
                last_trace = last_result[rank_var]
                nans = np.empty(child_shape)
                nans.fill(np.nan)
                last_result[rank_var] = np.concatenate([last_trace, nans])
            for rank_var in in_child - in_last:
                child_trace = child_result[rank_var]
                nans = np.empty(last_shape)
                nans.fill(np.nan)
                last_result[rank_var] = np.concatenate([nans, child_trace])
        return last_result


def _trace_node(node, workload: Workload, last_result):
    if isinstance(node, Spatial):
        raise NotImplementedError("Does not handle spatial for now.")
    elif isinstance(node, Temporal):
        rank_var = node.rank_variable
        tile_pattern = node.tile_pattern
        tile_shape = tile_pattern.tile_shape
        initial_tile_shape = tile_pattern.initial_tile_shape
        if initial_tile_shape is not None and initial_tile_shape != tile_shape:
            raise NotImplementedError("Does not handle imperfect for now")
        n_iterations = int(tile_pattern.calculated_n_iterations)
        last_shape = next(iter(last_result.values())).shape[0]
        next_result = {
            rank_var: np.tile(last_trace, n_iterations)
            for rank_var, last_trace in last_result.items()
        }
        next_result[rank_var] += np.repeat(
            np.arange(n_iterations) * tile_shape, last_shape
        )
        return next_result
    elif isinstance(node, Compute):
        einsum = workload.einsums[node.einsum]
        return {rank_var: np.ones((1,)) for rank_var in einsum.rank_variables}
    else:
        return last_result



from collections import defaultdict

from accelforge.frontend import arch
from accelforge.frontend.arch import Leaf, Memory, TensorHolder, Component
from accelforge.frontend.mapping import Compute, Mapping
from accelforge.frontend.spec import Spec

from accelforge.model._looptree.accesses import isl_buffer_accesses_from_buffet_actions
from accelforge.model._looptree.mapping_utilities import get_leaves
from accelforge.model._looptree.reuse.isl import IslReuseAnalysisOutput
from accelforge.model._looptree.reuse import SymbolicAnalysisOutput
from accelforge.model._looptree.types import Buffet

from accelforge.model._looptree.reuse.symbolic import BuffetStats
from accelforge.util._eval_expressions import MATH_FUNCS, eval_expression
from accelforge.util._sympy.broadcast_max import Max, Min
import sympy as sp


def isl_to_summarized(
    looptree_results: IslReuseAnalysisOutput, mapping, workload
) -> SymbolicAnalysisOutput:
    accesses_stats = isl_buffer_accesses_from_buffet_actions(
        looptree_results, mapping, workload, is_path=False
    )
    buffet_stats = {
        Buffet(level=component, tensor=tensor, einsum=einsum): BuffetStats(
            max_per_unit_read_actions=accesses.max_per_unit_reads,
            max_per_unit_write_actions=accesses.max_per_unit_writes,
        )
        for (component, tensor, einsum), accesses in accesses_stats.items()
    }
    return SymbolicAnalysisOutput(buffet_stats=buffet_stats)


def component_latency(
    looptree_results: SymbolicAnalysisOutput,
    flattened_arch: list[Leaf],
    mapping: Mapping,
    spec: Spec,
):
    component_to_actions: dict[str, dict[str, float]] = defaultdict(
        lambda: defaultdict(lambda: 0)
    )
    name2component: dict[str, Component] = {node.name: node for node in flattened_arch}

    compute_obj = flattened_arch[-1]
    if not isinstance(compute_obj, arch.Compute):
        raise ValueError("Last node in flattened_arch must be a Compute")

    for buffet, buffet_stats in looptree_results.buffet_stats.items():
        component = buffet.level
        actions = component_to_actions[component]
        if component not in name2component:
            raise ValueError(f"Component {component} found in mapping but not arch")

        for action in name2component[component].actions:
            actions[f"{action.name}_actions"] += 0

        if isinstance(name2component[component], TensorHolder):
            actions["read_actions"] += (
                buffet_stats.max_per_unit_read_actions
                - buffet_stats.min_per_unit_skipped_first_read_actions
            )
            if not isinstance(name2component[component], arch.Toll):
                actions["write_actions"] += (
                    buffet_stats.max_per_unit_write_actions
                    - buffet_stats.min_per_unit_skipped_first_write_actions
                )
        elif isinstance(name2component[component], arch.Compute):
            pass
        else:
            raise NotImplementedError(
                f"Component {component} is not a TensorHolder or Compute"
            )

    longest_compute_latency = Max(
        0, *[s.max_latency for s in looptree_results.compute_stats.values()]
    )
    component_to_actions[compute_obj.name]["compute_actions"] = longest_compute_latency

    # TODO: Unhardcode "compute" name"
    component_to_action_latency = defaultdict(dict)
    for component, actions in component_to_actions.items():
        component_obj = name2component[component]
        for action, count in actions.items():
            action_name = action.rsplit("_", 1)[0]
            latency = component_obj.actions[action_name].latency
            component_to_action_latency[component][f"{action_name}_latency"] = (
                latency * count
            )

    component_latency = {}

    symbol_table_base = {
        **dict(spec.variables),
        "variables": spec.variables,
        "max": Max,
        "min": Min,
        "sum": sp.Add,
    }

    for component, actions in component_to_actions.items():
        component_obj = name2component[component]
        symbol_table = {
            "action2latency": component_to_action_latency[component],
            **symbol_table_base,
            **dict(name2component[component]),
            **actions,
            **component_to_action_latency[component],
        }
        if name2component[component].total_latency is not None:
            component_latency[component] = eval_expression(
                name2component[component].total_latency,
                symbol_table,
                attr_name="latency",
                location=component,
            )

    return component_latency



import copy
import math
from accelforge._accelerated_imports import pandas as pd
from uuid import UUID
from collections import defaultdict

import sympy
from accelforge.frontend.mapping import Loop, Mapping, Spatial, Temporal
from accelforge.frontend.workload import EinsumName
from accelforge.util._frozenset import oset
from accelforge.mapper.FFM._join_pmappings.compatibility import (
    Compatibility,
)
from accelforge.mapper.FFM._join_pmappings.pmapping_dataframe import (
    MAPPING_COLUMN,
    PmappingDataframe,
    col2reservation,
    col_used_in_pareto,
    is_reservation_col,
    makepareto,
    tensor2col,
    reservation2col,
)

from accelforge.frontend.mapper.metrics import Metrics
from accelforge.mapper.FFM._make_pmappings.make_pmappings_from_templates.make_tile_shapes import (
    make_tile_shapes,
    IMPERFECT,
)
from accelforge.mapper.FFM._join_pmappings.pmapping_group import PmappingGroup
from accelforge.mapper.FFM._make_pmappings.pmapper_job import (
    Job,
    SameCompatibilityJobs,
)
from accelforge.mapper.FFM._pareto_df.df_convention import (
    is_fused_loop_col,
    is_n_iterations_col,
)
from accelforge.util._mathfuncs import _count_factorizations


def shift_reservations_by_null_loop_indices(
    mappings: pd.DataFrame, null_loop_indices: set[int]
):
    prev = copy.deepcopy(mappings)  # TODO: Is this needed?
    target2newabovename = {}
    dropcols = []
    for c in mappings.columns:
        if not is_reservation_col(c):
            continue
        reservation = col2reservation(c)
        name = reservation.name
        above = reservation.nloops
        new_above = above - sum(above > i for i in null_loop_indices)
        target = reservation2col(name, new_above)
        if target in target2newabovename:
            if above > target2newabovename[target][1]:
                dropcols.append(reservation2col(*target2newabovename[target]))
                target2newabovename[target] = (name, above)
            else:
                dropcols.append(c)
        else:
            target2newabovename[target] = (name, above)

    mappings.drop(columns=dropcols, inplace=True)
    renames = {}
    for target, (name, above) in target2newabovename.items():
        renames[reservation2col(name, above)] = target
    mappings.rename(columns=renames, inplace=True)
    if len(mappings.columns) != len(mappings.columns.unique()):
        shift_reservations_by_null_loop_indices(prev, null_loop_indices)
        raise ValueError(f"Duplicate columns: {mappings.columns}")
    return mappings


def get_equivalent_pmappings(
    pmapping_group: PmappingGroup, reservation_levels: set[int]
) -> list[PmappingGroup]:
    equivalent_permutations = pmapping_group.compatibility.make_equivalent_permutations(
        reservation_levels
    )
    result = [PmappingGroup(c, None) for c in equivalent_permutations]
    return result


def mapping2fused_loop_cols(mapping: Mapping, einsum_name: EinsumName):
    cols = []
    for loop in [l for l in mapping.nodes if isinstance(l, Loop) and l._fused]:
        if loop.tile_shape is not None:
            cols.append(loop.tile_shape)
        elif loop.tile_pattern is not None:
            cols.append(loop.tile_pattern.tile_shape)
            cols.append(loop.tile_pattern.initial_tile_shape)
        else:
            raise ValueError(f"Can't find tile shape or tile pattern for loop {loop}")
        return [f"{einsum_name}<SEP>{c}" if isinstance(c, str) else c for c in cols]


def get_fused_loop_indices(
    df: pd.DataFrame,
    compatibility: Compatibility,
    einsum_name: EinsumName,
    return_as_int: bool = False,
) -> pd.Series | int:
    result = []

    loops = compatibility.loops
    for i, loop in enumerate(loops):
        col = loop.tile_pattern.calculated_n_iterations
        assert col is not None, f"Loop {loop} has no calculated n_iterations"
        if isinstance(col, str):
            col = df[f"{einsum_name}<SEP>{col}"]
        elif isinstance(col, sympy.Symbol):
            col = df[f"{einsum_name}<SEP>{col.name}"]
        result.append(col != 1)

    if return_as_int:
        n = 0
        for b in result:
            n = n * 2 + b
        return n
    else:
        r2 = []
        for b in result:
            if len(b.unique()) > 1:
                raise ValueError(f"This won't work if there's more than one")
            r2.append(b.iloc[0])
        return r2


def _count_loops(job: Job) -> tuple[list[int], list[int], dict[str, int]]:
    nodes = job.mapping.nodes
    temporal_n_loops = []
    spatial_n_loops = []
    rv_spatial_count = defaultdict(int)
    rv_temporal_count = defaultdict(int)
    cur_n_loops = 0
    spatial_dim = None

    def pop_loop():
        nonlocal cur_n_loops
        if cur_n_loops >= 1:
            if spatial_dim is not None:
                spatial_n_loops.append(cur_n_loops)
            else:
                temporal_n_loops.append(cur_n_loops)
        cur_n_loops = 0

    for node in nodes:
        cur_spatial_dim = None
        if isinstance(node, Spatial):
            cur_spatial_dim = node.name
            rv_spatial_count[node.rank_variable] += 1
        if cur_spatial_dim != spatial_dim:
            pop_loop()
            spatial_dim = cur_spatial_dim
        if isinstance(node, Loop):
            cur_n_loops += 1
            if isinstance(node, Temporal):
                rv_temporal_count[node.rank_variable] += 1
        else:
            pop_loop()
    pop_loop()
    return temporal_n_loops, spatial_n_loops, rv_spatial_count, rv_temporal_count


def multiply_n_pmappings_by_permutations(n_pmappings: int, job: Job) -> int:
    option = job.spec_one_einsum.mapper._count_option_for_mapsapce_size_evaluation
    # if option == "normal":
    #     return n_pmappings

    temporal_n_loops, spatial_n_loops, rv_spatial_count, rv_temporal_count = (
        _count_loops(job)
    )

    rv = {k: v for k, v in job.rank_variable_bounds.items()}

    if "non_helpful_tile_shapes" in option:
        rv_temporal_count = {r: len(temporal_n_loops) for r in rv.keys()}

    if "non_helpful_loops_for_loop_orders" in option:
        for i in range(len(temporal_n_loops)):
            temporal_n_loops[i] = len(rv)

    # Count number of tile shapes
    rv2loops = {r: rv_spatial_count[r] + rv_temporal_count[r] for r in rv}
    n_factorizations = math.prod(
        _count_factorizations(b, rv2loops[r], imperfect=IMPERFECT)
        for r, b in rv.items()
    )
    n_temporal_loop_orders = math.prod(math.factorial(n) for n in temporal_n_loops)

    n = n_factorizations

    # assert n >= n_pmappings, f"n_pmappings: {n_pmappings} > n: {n}"

    if "redundant_loop_orders" in option:
        # job.mapping._n_loop_orders is the number of permutations that we actually
        # evaluate. Don't want to double count them.
        n *= n_temporal_loop_orders / job.mapping._n_loop_orders

    # assert n >= n_pmappings, f"n_pmappings: {n_pmappings} > n: {n}"

    return n


def assert_all_jobs_have_same_symbols(
    jobs_with_similar_compatibilities: SameCompatibilityJobs,
):
    iteration2symbols = []
    for j in jobs_with_similar_compatibilities:
        for t in j.compatibility.tensors:
            for i, l in enumerate(t.loops):
                if len(iteration2symbols) <= i:
                    iteration2symbols.append(oset())
                iteration2symbols[i].add(l.tile_pattern.calculated_n_iterations)
    assert all(
        len(s) == 1 for s in iteration2symbols
    ), "All jobs must have the same symbols for compatibility n_iterations"


def make_pmappings_from_templates(
    jobs_with_similar_compatibilities: SameCompatibilityJobs,
) -> tuple[EinsumName, list[PmappingGroup], dict[UUID, Mapping]]:
    jwsc = jobs_with_similar_compatibilities

    results = []
    pmapping_keep_rates = []

    for job in jobs_with_similar_compatibilities:
        try:
            result, tensor2mapping = make_tile_shapes(job)
        except Exception as e:
            e.add_note(f"Einsum {jwsc.einsum_name} compatibility {job.compatibility}")
            raise
        job.compatibility = job.compatibility.populate_loops(
            job.ranks_with_tile_pattern
        )

        # Ctrl-F for CONTIGUOUS_ITERATION_SPACE_DISCUSSION TODO: Turn tensor2pmapping
        # into per-tensor compatibility

        # This changes the pmapping count to include superfluous permutations
        # TODO: Add a multiplier for the permutations that we include in the fusion
        # piece, which are NOT known to be superfluous

        # prev = job.spec_one_einsum.mapper._count_option_for_mapsapce_size_evaluation
        # job.spec_one_einsum.mapper._count_option_for_mapsapce_size_evaluation = "redundant_loop_orders_and_irrelevant_loops"
        # a = multiply_n_pmappings_by_permutations(job.n_total_pmappings, job)
        # job.spec_one_einsum.mapper._count_option_for_mapsapce_size_evaluation = "redundant_loop_orders"
        # b = multiply_n_pmappings_by_permutations(job.n_total_pmappings, job)

        # if a < b:
        #     job.spec_one_einsum.mapper._count_option_for_mapsapce_size_evaluation = "redundant_loop_orders_and_irrelevant_loops"
        #     a = multiply_n_pmappings_by_permutations(job.n_total_pmappings, job)
        #     job.spec_one_einsum.mapper._count_option_for_mapsapce_size_evaluation = "redundant_loop_orders"
        #     b = multiply_n_pmappings_by_permutations(job.n_total_pmappings, job)
        #     assert False

        # job.spec_one_einsum.mapper._count_option_for_mapsapce_size_evaluation = prev
        job.n_total_pmappings = multiply_n_pmappings_by_permutations(
            job.n_total_pmappings, job
        )

        result[MAPPING_COLUMN] = job.job_id
        cols_to_drop = []
        for col in result.columns:
            if is_reservation_col(col):
                resource = col2reservation(col)[0]
                if resource in job.memories_track_pmappings_only:
                    cols_to_drop.append(col)
        result.drop(columns=cols_to_drop, inplace=True)
        results.append(result)
        pmapping_keep_rates.append(
            (job.job_id, dict(job.pmapping_keep_rates), job.n_total_pmappings)
        )

    fusable_tensors = jwsc.fusable_tensors
    einsum_name = jwsc.einsum_name
    metrics = jwsc.metrics
    drop_valid_reservations = not (Metrics.RESOURCE_USAGE & metrics)
    compatibility = jwsc.compatibility

    # Creating a PmappingDataframe fills in reservation columns since different pmappings
    # have different ones.
    next_shared_loop_index = compatibility.n_loops - 1
    df = PmappingDataframe.concat(
        [
            PmappingDataframe(
                r,
                skip_pareto=True,
                next_shared_loop_index=next_shared_loop_index,
                n_total_pmappings=1,  # Unused for now, just making an initial Pareto
                n_valid_pmappings=1,  # Unused for now, just making an initial Pareto
                ignored_resources=job.ignored_resources,
                # False because we may have lifetimes that stretch through this Einsum
                # due to data dependencies, not loops. For example, if we have no fused
                # loops, we can't free our reservations because another Einsum may
                # reserve a resource that lasts through us because it needs to propagate
                # to another Einsum. drop_valid_reservations is turned on later at the
                # joining stage because we have full information of live tensors then.
                drop_valid_reservations=False,
            )
            for r in results
        ],
        skip_pareto=True,
    ).data
    if df.empty:
        return einsum_name, [], {}, pmapping_keep_rates

    tensor_cols = [tensor2col(tensor) for tensor in fusable_tensors]
    df.columns = [
        c if col_used_in_pareto(c) or c in tensor_cols else f"{einsum_name}<SEP>{c}"
        for c in df.columns
    ]

    fused_loop_cols = [
        f"{einsum_name}<SEP>{c}"
        for c in compatibility.symbols()
        if not is_n_iterations_col(c)
    ]

    job0 = next(iter(jobs_with_similar_compatibilities))

    resource_usage_tolerance = job0.resource_usage_tolerance
    if not drop_valid_reservations:
        resource_usage_tolerance = job0.objective_tolerance

    # Absolute error can sum across Einsums, so we need to divide by it here. Relative
    # error doesn't (x * (1 += 0.1) + y * (1 += 0.1) = (x + y) * (1 +- <= 1.1)), so no
    # divide needed. NOTE: Pruning using tolerance happens in exactly two places that
    # affect results (a third is when there's dirty initial joining, but that doesn't
    # affect final results). Here and in the tile shape creation when a formula is fully
    # evaluated. There's no transformation of the values between these steps, so error
    # doesn't stack (it's just doing the same pruning, perhaps with a different number
    # of objectives).
    absolute_resource_usage_tolerance = (
        resource_usage_tolerance / job0.workload_n_einsums
    )

    # Pareto prune
    df = makepareto(
        df,
        split_by_cols=fused_loop_cols,
        resource_usage_tolerance=resource_usage_tolerance,
        absolute_resource_usage_tolerance=absolute_resource_usage_tolerance,
        objective_tolerance=job0.objective_tolerance,
    ).copy()

    jobs_passed_pareto = sorted(df[f"{einsum_name}<SEP>{MAPPING_COLUMN}"].unique())
    pmapping_objects = {
        job.job_id: job.mapping
        for job in jobs_with_similar_compatibilities
        if job.job_id in jobs_passed_pareto
    }

    assert_all_jobs_have_same_symbols(jobs_with_similar_compatibilities)
    # Otherwise, following logic fails

    df["fused_loop_indices"] = get_fused_loop_indices(
        df, job0.compatibility, einsum_name, return_as_int=True
    )
    groups = list(df.groupby(["fused_loop_indices"]))
    total_pmappings_per_group = sum(
        j.n_total_pmappings for j in jobs_with_similar_compatibilities
    ) / len(groups)
    valid_pmappings_per_group = sum(
        j.n_valid_pmappings for j in jobs_with_similar_compatibilities
    ) / len(groups)

    pmapping_groups = []
    for _, mappings in groups:
        compatibility = jwsc.compatibility
        fused_loop_indices = []

        for i, f in enumerate(
            get_fused_loop_indices(
                mappings, compatibility, einsum_name, return_as_int=False
            )
        ):
            if f:
                fused_loop_indices.append(i)

        null_loop_indices = tuple(
            i for i in range(compatibility.n_loops) if i not in fused_loop_indices
        )

        dropcols = ["fused_loop_indices"]
        mappings = mappings.drop(columns=dropcols)

        compatibility = compatibility.drop_loop_indices(null_loop_indices)

        symbol_renames, compatibility = compatibility.make_fused_loop_symbols(
            einsum_name
        )
        for k, v in symbol_renames.items():
            mappings[v] = mappings[f"{einsum_name}<SEP>{k}"]
        shift_reservations_by_null_loop_indices(mappings, null_loop_indices)

        energy_cols = [c for c in mappings.columns if "Total<SEP>energy" in c]
        if (mappings[energy_cols] < 0).any(axis=None):
            mapping_with_negative_energy = mappings[
                (mappings[energy_cols] < 0).any(axis=1)
            ]
            msg = ""
            for _, row in mapping_with_negative_energy.iterrows():
                for k, v in row.items():
                    msg += f"{k}: {v}\n"
                msg += "\n"
            raise RuntimeError(f"negative energy:\n{msg}")

        # Skip pareto because we already did it above
        next_shared_loop_index_this_group = compatibility.n_loops - 1
        mappings = compatibility.clear_unrelated_columns(mappings)
        partial_mappings = PmappingDataframe(
            mappings,
            next_shared_loop_index=next_shared_loop_index_this_group,
            n_total_pmappings=total_pmappings_per_group,
            n_valid_pmappings=valid_pmappings_per_group,
            ignored_resources=job.ignored_resources
            | oset(job.memories_track_pmappings_only),
            skip_pareto=True,
            # False because we may have lifetimes that stretch through this Einsum due
            # to data dependencies, not loops. For example, if we have no fused loops,
            # we can't free our reservations because another Einsum may reserve a
            # resource that lasts through us because it needs to propagate to another
            # Einsum. drop_valid_reservations is turned on later at the joining stage
            # because we have full information of live tensors then.
            drop_valid_reservations=False,
        )
        # If we have fewer fused loops, reservations likely got freed. We can free!
        if next_shared_loop_index_this_group != next_shared_loop_index:
            partial_mappings.make_pareto(
                resource_usage_tolerance=resource_usage_tolerance,
                objective_tolerance=job0.objective_tolerance,
                absolute_resource_usage_tolerance=absolute_resource_usage_tolerance,
            )
        pmapping_groups.append(PmappingGroup(compatibility, partial_mappings))

    # Defragment to speed up pickling
    for pg in pmapping_groups:
        pg.mappings._data = pg.mappings._data.copy()

    return (
        einsum_name,
        pmapping_groups,
        pmapping_objects,
        pmapping_keep_rates,
    )



from collections.abc import Mapping as MappingABC
import logging
from numbers import Number
from numbers import Real

from accelforge.frontend import arch
from accelforge.frontend.spec import Spec
from accelforge.model._looptree.reuse.symbolic import SymbolicAnalysisOutput
from accelforge.util._frozenset import oset
from accelforge.util._base_analysis_types import (
    ActionCount,
    ActionKey,
    VerboseActionKey,
)
from accelforge.model._looptree.types import Network


def gather_actions(
    looptree_results: SymbolicAnalysisOutput,
    bindings: dict[str, str],
    verbose: bool = False,
    use_name=False,
):
    actions: dict[tuple[str, str], ActionCount] = {}
    compute_levels = oset(c.level for c in looptree_results.compute_stats)

    buffet_keyer = _get_buffet_keyer(verbose, use_name, bindings)
    compute_keyer = _get_compute_keyer(verbose, use_name, bindings)
    network_keyer = _get_network_keyer(verbose, use_name, bindings)

    for buffet, accesses in looptree_results.buffet_stats.items():
        if buffet.level in compute_levels:
            continue

        level = buffet.level

        if use_name:
            level = level
        else:
            level = bindings[level]

        key = buffet_keyer(buffet, "read")
        if key not in actions:
            actions[key] = ActionCount.default()
        actions[key].total += accesses.net_total_read_actions()
        actions[key].max_per_unit += accesses.net_max_per_unit_read_actions()

        key = buffet_keyer(buffet, "write")
        if key not in actions:
            actions[key] = ActionCount.default()
        actions[key].total += accesses.net_total_write_actions()
        actions[key].max_per_unit += accesses.net_max_per_unit_write_actions()

    for compute, ops in looptree_results.compute_stats.items():
        key = compute_keyer(compute, "compute")
        if key not in actions:
            actions[key] = ActionCount.default()
        actions[key].total += ops.total_ops
        actions[key].max_per_unit += ops.max_per_unit_ops

    for network, stats in looptree_results.network_stats.items():
        key = network_keyer(network, "hops")
        if key not in actions:
            actions[key] = ActionCount.default()
        actions[key].total += stats.total_hops
        actions[key].max_per_unit += stats.max_hops

    return actions


def _get_buffet_keyer(verbose, use_name, bindings):
    if not verbose:

        def get_buffet_key(buffet, action_name) -> ActionKey:
            level = buffet.level
            if use_name:
                level = level
            else:
                level = bindings[level]
            return ActionKey(level, action_name)

    else:

        def get_buffet_key(buffet, action_name) -> VerboseActionKey:
            level = buffet.level
            if use_name:
                level = level
            else:
                level = bindings[level]
            return VerboseActionKey(level, action_name, buffet.tensor, buffet.einsum)

    return get_buffet_key


def _get_compute_keyer(verbose, use_name, bindings):
    if not verbose:

        def compute_keyer(compute, action_name):
            level = compute.level
            if use_name:
                level = level
            else:
                level = bindings[level]
            return ActionKey(level, action_name)

    else:

        def compute_keyer(compute, action_name):
            level = compute.level
            if use_name:
                level = level
            else:
                level = bindings[level]
            return VerboseActionKey(level, action_name, None, compute.einsum)

    return compute_keyer


def _get_network_keyer(verbose, use_name, bindings):
    if not verbose:

        def network_keyer(network: Network, action_name: str):
            component = network.component
            if not use_name:
                component = bindings[component]
            return ActionKey(component, action_name)

    else:

        def network_keyer(network: Network, action_name: str):
            component = network.component
            if not use_name:
                component = bindings[component]
            return VerboseActionKey(
                component, action_name, network.tensor, network.einsum
            )

    return network_keyer


def compute_energy_from_actions(
    spec: Spec,
    action_counts: MappingABC[ActionKey, Real],
    overall_latency: float,
    component_to_non_power_gated_porp: dict[str, int] = None,
) -> dict[ActionKey | VerboseActionKey, Number]:
    if component_to_non_power_gated_porp is None:
        logging.warning(
            "No component_to_non_power_gated_porp provided, will not account for power gating."
        )
        component_to_non_power_gated_porp = {}

    energy_result = {}
    components = {}
    for key, counts in action_counts.items():
        if counts.total == 0:
            continue
        if key.level not in components:
            components[key.level] = spec.arch.find(key.level)
        component_obj = components[key.level]
        try:
            energy_per_ac = component_obj.actions[key.action].energy
        except KeyError as e:
            raise KeyError(
                f"Action {key.action} not found in component {key.level}. Action occurred "
                f"{counts.total} times."
            ) from None
        energy_result[key] = counts.total * energy_per_ac

    for component_obj in spec.arch.get_nodes_of_type(arch.Component):
        energy_result[ActionKey(component_obj.name, "leak")] = (
            component_obj.total_leak_power
            * overall_latency
            * component_to_non_power_gated_porp.get(component_obj.name, 1)
        )

    return energy_result



from collections import defaultdict
from functools import cached_property
from typing import Any, Callable, Iterable
from accelforge._accelerated_imports import pandas as pd
from joblib import delayed

from accelforge.mapper.FFM._join_pmappings.pmapping_dataframe import PmappingDataframe

from accelforge.mapper.FFM._join_pmappings.compatibility import *
from accelforge.mapper.FFM._pareto_df.df_convention import (
    is_fused_loop_col,
    make_fused_loop_col,
)
from accelforge.util import parallel, oset


class PmappingGroup:
    def __init__(self, compatibility: Compatibility, mappings: PmappingDataframe):
        self.compatibility: Compatibility = compatibility
        self.mappings: PmappingDataframe = mappings
        self.tensors: dict[str, TensorReservation] = {
            t.name: t for t in self.compatibility.tensors
        }
        self.n_pre_prune_mappings = 0

        if isinstance(self.mappings, PmappingDataframe):
            checked = oset()
            for s in self.compatibility.symbols():
                checked.add(s)
                assert (
                    s in self.mappings.data.columns
                ), f"Column {s} not found in mappings"

            for col_name in self.mappings.data.columns:
                if col_name not in checked and is_fused_loop_col(col_name):
                    raise ValueError(f"Column {col_name} not found in compatibility")

    def compatibility_str(self):
        compatibility = ",".join(str(l) for l in self.compatibility.tensors)
        compatibility += " || " + ", ".join(str(t) for t in self.tensors.values())
        return compatibility

    @cached_property
    def tensor_names(self) -> set[str]:
        return oset(self.tensors)

    def copy(self) -> "PmappingGroup":
        return PmappingGroup(self.compatibility, self.mappings.copy())

    def __len__(self) -> int:
        return len(self.mappings)

    def merge_next(
        self,
        right: "PmappingGroup",
        live_tensors_post_join: set[str],
        live_tensors_with_right: set[str],
        aliased_tensors: dict[str, set[str]],
        compatibility_joined: Compatibility,
        ignored_resources: set[str],
        permuted_compatibility_left: Compatibility,
        permuted_compatibility_right: Compatibility,
        delay: bool = False,
        _pmapping_row_filter_function: Callable[[pd.Series], bool] | None = None,
        _force_allow_invalid_only_for_runtime_test: bool = False,
        _is_invalid: bool = False,
    ) -> "PmappingGroup":
        shared_loop_index = self.compatibility.shared_loop_index(
            right.compatibility.tensor_names | live_tensors_post_join
        )
        assert (
            shared_loop_index == self.compatibility.n_loops - 1
        ), "shared loop index not equal to left pmapping n_loops - 1"
        next_shared_loop_index = compatibility_joined.shared_loop_index(
            live_tensors_post_join
        )
        assert (
            next_shared_loop_index == compatibility_joined.n_loops - 1
        ), "next shared loop index not equal to joined pmapping n_loops - 1"
        assert compatibility_joined.tensor_names.issubset(
            live_tensors_post_join
        ), "joined compatibility includes tensors not live after joining"

        still_live_reservations = [
            r
            for r in self.compatibility.tensors
            if r.name in live_tensors_post_join
            and r.name not in right.compatibility.tensor_names
        ]

        duplicated_aliased_tensors = oset()
        for name, my_tensor in self.tensors.items():
            for aliased_tensor in aliased_tensors.get(name, oset()):
                if (aliased_tensor := right.tensors.get(aliased_tensor, None)) is None:
                    continue
                if my_tensor.resource_name == aliased_tensor.resource_name:
                    duplicated_aliased_tensors.add(aliased_tensor.name)

        mapping = delayed(self.mappings.merge_next)(
            right.mappings,
            duplicated_aliased_tensors,
            compatibility_left=permuted_compatibility_left,
            compatibility_right=permuted_compatibility_right,
            compatibility_joined=compatibility_joined,
            _pmapping_row_filter_function=_pmapping_row_filter_function,
            ignored_resources=ignored_resources,
            _force_allow_invalid_only_for_runtime_test=_force_allow_invalid_only_for_runtime_test,
            _is_invalid=_is_invalid,
        )

        if not delay:
            mapping = mapping[0](*mapping[1], **mapping[2])

        s = PmappingGroup(compatibility_joined, mapping)
        assert (
            compatibility_joined.max_above_loop_index == next_shared_loop_index + 1
        ), f"{self.compatibility} {right.compatibility} {next_shared_loop_index + 1} -> {compatibility_joined} {len(compatibility_joined.loops)}"
        s.tensors.update(right.tensors)
        s.tensors.update(self.tensors)
        s.n_pre_prune_mappings = len(self.mappings.data) * len(right.mappings.data)
        return s

    def get_shared_loop_index(self, live_tensors: set[str]) -> int:
        live_tensors = list(self.compatibility.tensor_names) + [live_tensors]
        return self.compatibility.shared_loop_index(live_tensors)

    def _right_consolidate(
        self,
        live_tensors: set[str] = None,
        shared_tensors: set[str] = None,
    ):
        dead_tensors = oset(self.tensors) - (live_tensors or oset())
        check_tensors = (shared_tensors or oset()) | (live_tensors or oset())
        shared_loop_index = self.compatibility.shared_loop_index(check_tensors)
        for t in dead_tensors:
            t = self.tensors.pop(t)
        if self.mappings.free_to_loop_index(shared_loop_index):
            self.mappings.make_pareto()
        return self

    def _left_consolidate(self, live_tensors: set[str] = None):
        check_tensors = live_tensors or oset()
        shared_loop_index = self.compatibility.shared_loop_index(check_tensors)
        self.mappings.free_to_loop_index(shared_loop_index)
        if live_tensors is None:
            self.mappings.clear_fused_loop_symbols()
        return self

    @staticmethod
    def right_consolidate(
        pmapping_groups: list["PmappingGroup"],
        live_tensors: set[str],
        shared_tensors: set[str] = None,
        pbar: str = None,
        parallelize: bool = True,
    ) -> list["PmappingGroup"]:
        def job(s):
            return s._right_consolidate(live_tensors, shared_tensors)

        if not parallelize:
            return [
                s._right_consolidate(live_tensors, shared_tensors)
                for s in pmapping_groups
            ]

        return parallel([delayed(job)(s) for s in pmapping_groups], pbar=pbar)

    @staticmethod
    def left_consolidate(
        pmapping_groups: list["PmappingGroup"],
        live_tensors: set[str],
        pbar: str = None,
        parallelize: bool = True,
    ) -> list["PmappingGroup"]:
        def job(s):
            return s._left_consolidate(live_tensors)

        if not parallelize:
            return [s._left_consolidate(live_tensors) for s in pmapping_groups]

        return parallel([delayed(job)(s) for s in pmapping_groups], pbar=pbar)

    def _hashable_attrs(self):
        return self.mappings, fzs(self.tensors.items())

    @staticmethod
    def concat(
        pmapping_groups: Iterable["PmappingGroup"],
        allow_different_compatibilies: bool = False,
    ) -> "PmappingGroup":
        pmapping_groups = list(pmapping_groups)
        assert len(pmapping_groups) > 0, "Cannot concat empty list of PmappingGroups"
        if not allow_different_compatibilies:
            s = oset(
                s.compatibility.clear_symbolic_tile_patterns() for s in pmapping_groups
            )
            if len(s) > 1:
                a = pmapping_groups[0]
                for b in pmapping_groups[1:]:
                    if a.compatibility != b.compatibility:
                        break
                assert (
                    a == b
                ), f"Cannot concat PmappingGroups with different compatibilies:\n\t{a}\n\t{b}"
                assert len(s) == 1, (
                    f"Cannot concat PmappingGroups with different compatibilies:\n\t"
                    + "\n\t".join(str(s2) for s2 in s)
                )

        c0 = pmapping_groups[0].compatibility
        to_concat = [pmapping_groups[0]] + [
            s.rename_compatibility(c0) for s in pmapping_groups[1:]
        ]
        catted = PmappingDataframe.concat([s.mappings for s in to_concat])
        return PmappingGroup(c0, catted)

    def rename_compatibility(self, new_c: Compatibility) -> Compatibility:
        c, renamed = self.compatibility._rename_to_match(new_c)
        return PmappingGroup(c, self.mappings.rename(renamed))

    @staticmethod
    def _group(
        pmapping_groups: list["PmappingGroup"],
        live_tensors: set[str] | Literal["All"],
        clear_tile_patterns_and_reservation_indices: bool = False,
        include_permutations: bool = False,
        clear_symbolic_tile_patterns: bool = False,
        try_permute_into_equivalent: bool = False,
        # mixable_ranks: dict[Rank, set[Rank]] = None,
    ) -> (
        dict[Compatibility, list["PmappingGroup"]]
        | dict[Compatibility, list[tuple["PmappingGroup", list[int]]]]
    ):
        """
        Clears dead tensors (may keep loops), then group PmappingGroups based on
        compatibility.
        """
        grouped = defaultdict(list)

        def clear(c: Compatibility):
            if clear_symbolic_tile_patterns:
                c = c.clear_symbolic_tile_patterns()
            if clear_tile_patterns_and_reservation_indices:
                return c.clear_tile_patterns_and_reservation_indices()
            return c

        for s in pmapping_groups:
            compatibility = s.compatibility.clear_dead_tensors(live_tensors)

            if include_permutations or try_permute_into_equivalent:
                keys = compatibility.make_equivalent_permutations()
                for t, loop_changes in keys:
                    # Line below DOES NOT MUTATE. It's check that the permutation works.
                    s.compatibility.permute(loop_changes)
                    grouped[clear(t)].append((s, loop_changes))
            else:
                grouped[clear(compatibility)].append(s)

        if clear_tile_patterns_and_reservation_indices:
            for k in grouped:
                assert (
                    len(k.reservation_indices) == 0
                ), f"Extra reservation indices are not empty: {k.reservation_indices}"

        # if mixable_ranks is not None:
        #     new_grouped = {}
        #     for c, g in grouped.items():
        #         for c2 in c.get_equivalent_compatibilities(mixable_ranks):
        #             new_grouped.setdefault(c2, []).extend(g)
        #     grouped = new_grouped

        if try_permute_into_equivalent:
            assert not include_permutations
            new_grouped = {}
            pmgroups_remaining = oset(id(s) for s in pmapping_groups)
            for c, g in sorted(grouped.items(), key=lambda x: len(x[1]), reverse=True):
                if not pmgroups_remaining:
                    break
                g = [
                    (s, loop_changes)
                    for s, loop_changes in g
                    if id(s) in pmgroups_remaining
                ]
                if g:
                    pmgroups_remaining -= oset(id(s) for s, _ in g)
                    permuted = [
                        PmappingGroup(
                            s.compatibility.permute(lc),
                            s.mappings.clear_irrelevant_columns(s.compatibility),
                        )
                        for s, lc in g
                    ]
                    new_grouped[c] = permuted
            grouped = new_grouped

        return grouped

    @staticmethod
    def combine_combineable(
        pmapping_groups: list["PmappingGroup"],
        live_tensors: set[str] | Literal["All"],
        allow_different_compatibilies: bool = False,
        _combine_reservations: bool = True,
        print_progress: bool = True,
        pbar_postfix: str = "",
    ) -> list["PmappingGroup"]:
        pmapping_groups = [s for s in pmapping_groups if len(s.mappings.data) > 0]
        no_combine = []
        if not _combine_reservations:
            has_reservations = [s.mappings.has_reservations() for s in pmapping_groups]
            no_combine = [s for s, h in zip(pmapping_groups, has_reservations) if h]
            pmapping_groups = [
                s for s, h in zip(pmapping_groups, has_reservations) if not h
            ]
        groups = list(
            PmappingGroup._group(
                pmapping_groups,
                live_tensors,
                clear_symbolic_tile_patterns=True,
                try_permute_into_equivalent=True,
            ).values()
        )
        groups_with_one = [g[0] for g in groups if len(g) == 1]
        if len(groups_with_one) == len(groups):
            return groups_with_one + no_combine

        others = parallel(
            [
                delayed(PmappingGroup.concat)(g, allow_different_compatibilies)
                for g in groups
                if len(g) > 1
            ],
            pbar=f"Grouping pmappings{pbar_postfix}" if print_progress else None,
        )
        return groups_with_one + others + no_combine

    @staticmethod
    def filter_by_tensors(
        pmapping_groups: list["PmappingGroup"] | dict[Compatibility, Any],
        tensors: set[str],
    ) -> list["PmappingGroup"]:
        def check(tensors_to_check):
            for t in tensors_to_check:
                for t2 in tensors:
                    if (t2.name == "*" or t.name == t2.name) and t != t2:
                        return False
            return True

        tensors = oset(tensors)
        if isinstance(pmapping_groups, list):
            return [s for s in pmapping_groups if check(s.compatibility.tensors)]
        if isinstance(pmapping_groups, dict):
            return {k: v for k, v in pmapping_groups.items() if check(k.tensors)}
        raise ValueError(f"Invalid type {type(pmapping_groups)}")

    @staticmethod
    def group(
        pmapping_groups: list["PmappingGroup"], live_tensors: set[str]
    ) -> dict[tuple[Compatibility, ...], list[tuple["PmappingGroup", list[int]]]]:
        x = PmappingGroup._group(
            pmapping_groups,
            live_tensors,
            clear_tile_patterns_and_reservation_indices=True,
            include_permutations=True,
            # mixable_ranks=mixable_ranks,
        )
        return x

    @staticmethod
    def remove_dead_tensors(
        pmapping_groups: list["PmappingGroup"], live_tensors: set[str]
    ):
        for s in pmapping_groups:
            for t in list(s.tensors):
                if t not in live_tensors:
                    del s.tensors[t]



"""Timeloop Spec. Each piece below (minus processors) corresponds to a top key in the Timeloop spec."""

from .spec import *
import accelforge.frontend.arch as arch
import accelforge.frontend.config as config
import accelforge.frontend.mapping as mapping
import accelforge.frontend.renames as renames
import accelforge.frontend.spec as spec
import accelforge.frontend.variables as variables
import accelforge.frontend.workload as workload

__all__ = [
    "arch",
    "config",
    "mapping",
    "renames",
    "spec",
    "variables",
    "workload",
    "Spec",
]



from typing import Any, Annotated, Literal

from accelforge.frontend.mapper.metrics import Metrics
from accelforge.frontend.renames import EinsumName
from accelforge.util._basetypes import EvalableModel


class FFM(EvalableModel):
    """Configuration for the Fast and Fusiest Mapper."""

    metrics: Metrics = Metrics.ENERGY
    """ Metrics used to optimize mappings. """

    info_metrics: Metrics = Metrics.all_metrics()
    """Metrics to be reported for final mappings."""

    _timeloop_style_even: bool = False
    """ Timeloop-style even mappings must have, for each memory, at most two
        locations where storage nodes may be placed. """

    force_memory_hierarchy_order: bool = True
    """
    If set to true, storage nodes for lower-level memories must be placed below storage
    nodes for higher-level memories. For example, all MainMemory storage nodes must go
    above all LocalBuffer storage nodes.

    This constraint always applies to same-tensor storage nodes (e.g., MainMemory
    reusing Output must go above LocalBuffer reusing Output); turning it off will
    permit things like MainMemory reusing Output going above LocalBuffer reusing
    Input.
    """

    out_of_order_hierarchy_explore_removing_spatials_for_more_temporals: bool = False
    """
    If force_memory_hierarchy_order is set to False or is set to False for any
    particular component, and a spatial fanout ends up being raised above a storage node
    that does not have that fanout, then there may be cases where a spatial loop is put
    above a component that does not have the associated fanout.

    When this happens, we may not put between the spatial and the storage node any
    temporal loops that affect the same indexing expressions as the spatial loops.

    For example, the following is not allowed:

    Arch:

    - Global Buffer
    - 2x fanout
    - Register

    Mapping:

    spatial-for-reg n in [0, 10):
      [Register reuses input]
        for n in [0, 2):
          [Global Buffer reuses output]

    By default, if there are spatial loops that are not constrained away, then the
    mapper will not explore putting any temporal loops that conflict. In the above
    example, it will never place the above temporal loop. If this is set to True, then
    the mapper will explore removing the spatial loop in order to allow for the temporal
    loop to be placed. In the above example, it will explore removing the spatial loop
    in order to allow for the temporal loop to be placed.
    """

    max_fused_loops_per_rank_variable: int = 1
    """ The maximum number of fused loops in a pmapping for a given rank variable. """

    max_fused_loops: float | int = float("inf")
    """ The maximum total number of fused loops in a pmapping. """

    max_loops: float | int = float("inf")
    """ The maximum total loops in a pmapping. """

    max_loops_minus_ranks: float | int = float("inf")
    """
    The maximum total loops in a pmapping minus the number of ranks. For example,
    3 means that the number of loops can be up to (the number of ranks + 3).
    """

    explore_loop_orders: bool = True
    """
    Whether to explore loop orders for loops where we may get partial reuse. Note that
    loop orders that don't matter (i.e., ones that have either full or no reuse) are not
    explored, except in joining where we may join partial mappings who have
    different-but-equivalent loop orders.
    """

    _can_lower_outermost_memory: bool = False
    """
    Whether the storage node of outermost memory can be lowered. If set to True, the
    mapper may exchange tiles of tensors via the outermost memory, instead of storing
    full tensors. Set this to True to explore reducing outermost memory usage.

    TODO: Also need to explore putting loops above the outermost memory then. This is
    currently private because we may want to have a catch-all term like
    "save_outermost_memory_usage".
    """

    _only_output_pmapping_with_index: (
        int | set[int] | dict[EinsumName, int | set[int]] | None
    ) = None
    """
    For debugging. Only output the pmapping with this index. If a dictionary, then the
    keys are einsum names and the values are the indices.
    """

    memory_limit: float | int = float("inf")
    """ The maximum memory limit for the mapper. """

    memory_limit_per_process: float | int = float("inf")
    """ The maximum memory limit per process for one of the mapper's processes. """

    time_limit: float | int = float("inf")
    """ The maximum time limit for the mapper. """

    time_limit_per_pmapping_template: float | int = float("inf")
    """ The maximum time limit per pmapping template. """

    max_pmapping_templates_per_einsum: float | int = float("inf")
    """
    The maximum number of pmapping templates per Einsum. Once this many templates are
    generated, the mapper will stop generating more. This is useful for debugging (why
    are so many templates being generated?).
    """

    prioritize_reuse_of_unfused_tensors: bool = False
    """
    If set to True, then for all memory levels, the mapper will place the storage nodes
    of unfused tensors above those of fused tensors. This is overridden if there is any
    tensor_order_options specified for a memory level. The result of this is that the
    mapper will avoid mappings that repeatedly fetch unfused tensors in order to allow
    for smaller tiles of fused tensors. This may lead to better mappings,
    but slows down the mapper.
    """

    _count_option_for_mapsapce_size_evaluation: tuple[
        Literal[
            "redundant_loop_orders",
            "non_helpful_loops_for_loop_orders",
            "non_helpful_tile_shapes",
            "redundant_dataplacements",
        ]
    ] = ()

    objective_tolerance: float = 0
    """
    Reduces memory usage and runtime for the mapper. When set to a nonzero value, the
    mapper may return mappings up to (1 + tolerance)× optimal. Also see
    resource_usage_tolerance to further reduce mapper memory usage and runtime.
    """

    resource_usage_tolerance: float = 0
    """
    Reduces memory usage and runtime for the mapper. When set to a nonzero value, the
    mapper may drop mappings with resource usage > (1 - tolerance)× optimal. The mapper
    is guaranteed to return all Pareto-optimal mappings with resource usage below this,
    and perhaps more. If Metrics.RESOURCE_USAGE is set, then this is ignored. Setting
    this, as well as objective_tolerance, to a greater-than-zero value will reduce
    memory usage for the mapper.
    """

    _let_non_intermediate_tensors_respawn_in_backing_storage: bool = False
    """
    If set to True, we can have temporal loops above the backing storage for
    non-intermediate tensors, which effectively causes them to respawn.
    """

    _skip_invalid: bool = True
    """
    Whether to skip invalid joinings. This is used for a paper ablation study. Do not
    use this unless you're ablating or want to burn CPU cycles.
    """

    _combine_reservations: bool = True
    """
    Whether to combine reservations to increase pruning effectiveness. This is used for
    a paper ablation study. Do not use this unless you're ablating or want to burn CPU
    cycles.
    """

    _runtime_log_file: str | None = None
    """
    If set, append per-step runtime as JSON lines to this file. Used for ablation study
    measurements.
    """

    _metric_aggregator: Literal["sum", "prod", "any"] = "any"
    """
    How to aggregate metrics together to determine whether one pmapping is better than
    another. "sum" means that the metrics are added together, "prod" means that the
    metrics are multiplied together, and "any" means that any metric being better than
    the other is enough to consider it non-dominated.
    """



from copy import copy, deepcopy
from uuid import uuid4

from accelforge._accelerated_imports import pandas as pd

from accelforge.frontend import arch
from accelforge.frontend.arch import Memory
from accelforge.frontend.mapping.mapping import MappingNodeWithChildren
from accelforge.frontend.renames import EinsumName, TensorName
from accelforge.frontend.spec import Mapping, Spec
from accelforge.util._frozenset import oset
from accelforge.frontend.mapping import (
    Compute,
    Reservation,
    Split,
    Nested,
    NodeList,
    TensorHolder,
)
from accelforge.frontend.workload import Workload
from accelforge.frontend._workload_isl._symbolic import (
    get_stride_and_halo,
    get_rank_variable_relevancy,
)
from accelforge.mapper.FFM._make_pmappings.make_pmappings_from_templates.symbol_relations import (
    get_initial_delta_choices,
)
from accelforge.mapper.FFM._pareto_df.df_convention import col_used_in_joining


def evaluate_mapping(
    spec: Spec,
    flattened_arches: dict[(EinsumName, str), list[arch.Leaf]] | None = None,
    evaluated_specs: dict[EinsumName, Spec] | None = None,
):
    """
    Evaluate a mapping.

    Parameters
    ----------
    spec:
        The specification of architecture, workload, and mapping.
    flattened_arches:
        A dictionary of (EinsumName, Compute Name) to lists of architecture nodes. These
        contain the evaluated and flattened architecture node for that particular Einsum
        and compute combination. If provided, then these will be used instead of
        re-parsing the architecture.
    evaluated_specs:
        A dictionary of Einsum names to evaluated specifications. These contain the evaluated
        specification for that particular Einsum. If provided, then these will be used
        instead of re-parsing the specification.
    """
    from accelforge.mapper.FFM._join_pmappings.compatibility import Compatibility
    from accelforge.mapper.FFM._join_pmappings.pmapping_dataframe import (
        PmappingDataframe,
    )
    from accelforge.mapper.FFM._join_pmappings.pmapping_group import PmappingGroup
    from accelforge.mapper.FFM._join_pmappings.join_pmappings import (
        clean_compress_and_join_pmappings,
    )
    from accelforge.mapper.FFM.pmappings import MultiEinsumPmappings
    from accelforge.mapper.FFM._make_pmappings.make_pmappings import (
        get_rank_variable_bounds_for_all_einsums,
    )
    from accelforge.model.run_model import (
        run_model,
    )
    from accelforge.mapper.FFM._make_pmappings.make_pmappings_from_templates.make_tile_shapes import (
        _calculate_iterations_and_rank_columns,
        _clean_energy_columns,
    )
    from accelforge.mapper.FFM._make_pmappings.pmapper_job import Job

    assert (evaluated_specs is not None) == (
        flattened_arches is not None
    ), f"Provide either flattened_arches or evaluated_specs, not both."

    original_job = Job(
        metrics=spec.model.metrics,
        rank_variable_bounds=get_rank_variable_bounds_for_all_einsums(spec),
        spec_one_einsum=spec,
        resource_usage_tolerance=0,  # spec.model.resource_usage_tolerance,
        objective_tolerance=0,  # spec.model.objective_tolerance,
        workload_n_einsums=len(spec.workload.einsum_names),
    )

    einsum2pmappings = {}
    pmapping_objects = {}
    einsum2jobs = {}
    s = (
        "Spec must not be evaluated before evaluating a mapping. Was "
        "this spec returned by spec.calculate_component_area_energy_latency_leak()?"
    )

    needs_reservations = not bool(spec.mapping.get_nodes_of_type(Reservation))

    fusable_tensors = spec.workload.tensor_names_used_in_multiple_einsums
    stride_and_halo = get_stride_and_halo(spec.workload)

    assert not getattr(spec, "_evaluated", False), s
    for pmapping in _split_mapping_to_pmappings(spec.mapping, spec.workload):
        einsum_name = pmapping.nodes[-1].einsum
        compute_name = pmapping.nodes[-1].component
        pmapping_id = uuid4()
        job = copy(original_job)

        if flattened_arches is not None:
            flattened_arch = flattened_arches[(einsum_name, compute_name)]
            cur_spec = evaluated_specs[einsum_name]

        else:
            cur_spec = spec.calculate_component_area_energy_latency_leak(
                einsum_name=einsum_name,
                area=False,
            )
            flattened_arch = cur_spec._get_flattened_architecture(
                compute_node=pmapping.nodes[-1].component
            )

        job.spec_one_einsum = cur_spec
        job.einsum_name = pmapping.nodes[-1].einsum
        job.stride_and_halo = stride_and_halo
        # spec, not cur_spec, becuase cur_spec only has one einsum and the delta choices
        # depend on >1 Einsums
        job.initial_delta_choices = get_initial_delta_choices(
            job.einsum_name, spec.workload
        )
        pmapping.split_reservations()
        pmapping.split_loop_with_multiple_rank_variables(job.einsum_name)
        pmapping.split_tensor_holders_with_multiple_tensors()
        _add_backing_to_tensor_holders(pmapping)

        job.mapping = pmapping
        job.tensor_to_relevancy = {
            tensor: get_rank_variable_relevancy(
                job.spec_one_einsum.workload.einsums[job.einsum_name], tensor
            )
            for tensor in job.spec_one_einsum.workload.einsums[
                job.einsum_name
            ].tensor_names
        }
        pmapping.clear_irrelevant_reservations(oset(job.tensor_to_relevancy))

        einsum2jobs[job.einsum_name] = job

        job.flattened_arch = flattened_arch
        job.memories_track_all = [
            m.name for m in flattened_arch if isinstance(m, Memory)
        ]

        job.fusable_tensors = fusable_tensors & oset(job.tensor_to_relevancy)
        einsum = cur_spec.workload.einsums[job.einsum_name]

        _, df, _, _, tensor2mapping, _ = run_model(
            job, add_reservations=needs_reservations
        )

        # Calculate iteration counts and rank columns
        _clean_energy_columns(df, job.metrics)
        _calculate_iterations_and_rank_columns(
            job.mapping.nodes, job, df, job.rank_variable_bounds
        )
        compatibility = Compatibility.from_mapping(
            job.mapping,
            job.fusable_tensors,
            einsum,
        )
        symbol_renames, compatibility = compatibility.make_fused_loop_symbols(
            einsum_name
        )
        for k, v in symbol_renames.items():
            df[v] = df.pop(k)

        new_df = {}
        for key, value in df.items():
            if not col_used_in_joining(key):
                key = f"{job.einsum_name}<SEP>{key}"
            # Want usage both for joining & for per-einsum info
            if key.startswith("usage<SEP>"):
                new_df[f"{job.einsum_name}<SEP>{key}"] = value
            new_df[key] = value
        df = new_df
        df[f"{job.einsum_name}<SEP>mapping"] = pmapping_id

        einsum2pmappings[job.einsum_name] = [
            PmappingGroup(
                compatibility,
                PmappingDataframe(
                    data=pd.DataFrame(df, columns=df.keys(), index=[0]),
                    n_total_pmappings=1,
                    n_valid_pmappings=1,
                    ignored_resources=oset(),
                    drop_valid_reservations=False,
                ),
            )
        ]
        pmapping_objects[job.einsum_name] = {pmapping_id: job.mapping}

    # Restore the original order
    einsum2pmappings = {
        einsum_name: einsum2pmappings[einsum_name]
        for einsum_name in spec.workload.einsum_names
        if einsum_name in einsum2pmappings
    }

    return clean_compress_and_join_pmappings(
        pmappings=MultiEinsumPmappings(
            spec=spec,
            einsum2pmappings=einsum2pmappings,
            pmapping_objects=pmapping_objects,
            einsum2jobs=einsum2jobs,
            can_combine_multiple_runs=False,
            einsums_with_pmappings_generated=oset(spec.workload.einsum_names),
            flattened_arches=flattened_arches,
            evaluated_specs=evaluated_specs,
        ),
        metrics=spec.model.metrics,
        print_progress=False,
        for_model=True,
    )


def _add_backing_to_tensor_holders(pmapping: Mapping):
    seen_tensors = oset()
    for node in pmapping.nodes:
        if isinstance(node, TensorHolder):
            new_tensors = oset(node.tensors) - seen_tensors
            node._backing = new_tensors
            seen_tensors.update(new_tensors)


def _split_mapping_worker(node: MappingNodeWithChildren):
    if isinstance(node, Split):
        for subnodes in node.nodes:
            yield from _split_mapping_worker(subnodes)
        return

    assert isinstance(node, Nested), "BUG"

    for n in node.nodes[:-1]:
        assert not isinstance(n, MappingNodeWithChildren), "BUG"

    if not isinstance(node.nodes[-1], MappingNodeWithChildren):
        yield node.nodes
        return

    for subnodes in _split_mapping_worker(node.nodes[-1]):
        yield node.nodes[:-1] + subnodes


def _split_mapping_to_pmappings(mapping: Mapping, workload: Workload):
    """
    A DFS-like algorithm to split a mapping into pmappings at Split nodes.

    DFS has to be modified because the tree has list of nodes for nested nodes
    instead of links to children.
    """
    for nodes in _split_mapping_worker(mapping):
        mapping = Mapping(nodes=deepcopy(nodes))
        _remove_storage_of_unrelevant_tensors(mapping, workload)
        yield mapping


def _remove_storage_of_unrelevant_tensors(pmapping: Mapping, workload: Workload):
    """
    Remove tensors from Storage nodes that are not relevant to the Einsum being
    mapped.
    """
    einsum_name = pmapping.nodes[-1].einsum
    einsum = workload.einsums[einsum_name]
    relevant_tensors = oset(t.name for t in einsum.tensor_accesses)

    new_nodes = []
    for node in pmapping.nodes:
        if isinstance(node, TensorHolder):
            node.tensors = [t for t in node.tensors if t in relevant_tensors]
            if node.tensors:
                new_nodes.append(node)
        else:
            new_nodes.append(node)

    pmapping.nodes = new_nodes



from accelforge.frontend.mapping.mapping import MappingNode


import copy
from collections import defaultdict
import itertools
import logging
from typing import Any, Iterator, List
import uuid

from tqdm import tqdm

import accelforge.frontend.arch as arch
from accelforge.util._frozenset import oset
from accelforge.frontend.mapping import (
    Compute,
    Loop,
    Mapping,
    MappingNode,
    Spatial,
    TensorHolder,
    Temporal,
)
from accelforge.frontend.renames import Rank
from accelforge.frontend.spec import Spec
from accelforge.frontend._workload_isl._isl import get_rank_variable_bounds
from accelforge.frontend._workload_isl._symbolic import (
    Relevant,
    get_rank_variable_relevancy,
    get_stride_and_halo,
    get_stride_and_halo_of_einsum,
    PartiallyRelevant,
)
from accelforge.frontend.workload import (
    TensorName,
    Einsum,
    EinsumName,
    RankVariable,
    Workload,
    SymbolTable,
)
from accelforge.mapper.FFM._make_pmappings.make_pmapping_templates.make_storage_order import (
    get_tensor_choices,
)
from accelforge.mapper.FFM._make_pmappings.make_pmapping_templates.make_reservations import (
    get_reservation_choices,
)
from accelforge.mapper.FFM._make_pmappings.contraints.constraints import (
    MappingConstraints,
    get_constraints,
)
from accelforge.mapper.FFM._make_pmappings.make_pmapping_templates.make_loops import (
    insert_temporal_loops,
    insert_spatial_loops,
)
from accelforge.mapper.FFM._make_pmappings.pmapper_job import (
    Job,
    SameEinsumJobs,
)
from accelforge.mapper.FFM._join_pmappings.compatibility import Compatibility
from accelforge.model._looptree.reuse.symbolic import (
    label_fused_loops,
    quick_insert_reservation_nodes,
)


def unpack_loops_to_rank_variables(mapping: List[MappingNode]):
    mapping_new = []
    for node in mapping:
        if not isinstance(node, Loop) or not isinstance(node.rank_variable, set):
            mapping_new.append(node)
            continue

        for r in sorted(node.rank_variable):
            mapping_new.append(
                type(node)(
                    rank_variable=r,
                    **node.model_dump(exclude={"rank_variable"}, recursive=False),
                )
            )
    return mapping_new


# =================================================================================================
# Iterate over mappings
# =================================================================================================
def place_missing_temporal_loops(
    mapping: List[MappingNode], einsum: Einsum, flattened_arch: list[arch.Leaf]
):
    """
    Adds temporal loops to the mapping to fill in any rank variables that are missing.
    This may occur if there are no points where it'd be helpful to add a non-fused loop,
    so we just need to add one somewhere.
    """
    # If any rank variables are missing, add them as high as possible.

    rank_variables = einsum.rank_variables
    for m in mapping:
        if isinstance(m, Temporal) and not m._fused:
            rank_variables.discard(m.rank_variable)

    # Insert point: Right under the last backing & below any out-of-order fanouts
    fanouts = {}
    fanout = 1
    for node in flattened_arch:
        fanouts[node.name] = (fanout := fanout * node.get_fanout())

    insert_point = 0
    greatest_previous_fanout = 1
    for i in range(len(mapping)):
        if isinstance(mapping[i], TensorHolder):
            if mapping[i]._backing:
                insert_point = i + 1
            cur_fanout = fanouts[mapping[i].component]
            if cur_fanout < greatest_previous_fanout:
                insert_point = i + 1
            greatest_previous_fanout = max(greatest_previous_fanout, cur_fanout)

        # Put it below all the other temporals here in case we're lowering through them
        if isinstance(mapping[i], Temporal) and insert_point == i:
            insert_point = i + 1

    temporals = [Temporal(rank_variable=r) for r in sorted(rank_variables)]

    if insert_point == len(mapping):
        mapping.extend(temporals)
    else:
        for t in temporals:
            mapping.insert(insert_point, t)


def remove_unordered_spatial_temporal_loops(
    mapping: list[MappingNode],
    flattened_arch: list[arch.Leaf],
    einsum: Einsum,
    explore_unordered_spatial_loops: bool = True,
):
    fanout = 1
    fanouts = {}
    for node in flattened_arch:
        fanouts[node.name] = (fanout := fanout * node.get_fanout())

    # TensorHolders' tiles include the full iteration space of temporal loops below
    # them. If all of the following are true:
    # - Any later spatials have <= the fanout of this TensorHolder
    # - There's a temporal loop between the TensorHolder and the Spatial
    # - The temporal/spatial have the same rank variable
    # - The rank variable indexes into the tensor
    #
    # Then the temporal/spatial pair will result in a non-contiguous tile of the
    # iteration space, which is not supported and must be removed.

    tensor2rvs = einsum.tensor2rank_variables
    disallowed_combinations: list[tuple[set[int], set[int]]] = []
    for i, node in enumerate(mapping):
        # Track TensorHolders that have been seen and the rank variables that affect
        # them
        if not isinstance(node, TensorHolder):
            continue

        relevent_rvs = oset.union(*[tensor2rvs[t] for t in node.tensors])

        # Find the last spatial whose component's fanout is <= the TensorHolder's
        # component fanout. This spatial will affect the TensorHolder's tile.
        check_up_to = i
        for j, node2 in enumerate(mapping[i + 1 :]):
            if isinstance(node2, Spatial):
                if fanouts[node.component] >= fanouts[node2.component]:
                    check_up_to = i + j + 1

        # Find all temporal and spatial loops between the TensorHolder and the last
        # spatial that affects it.
        rv2spatial = {}
        rv2temporal = {}
        for node2 in mapping[i + 1 : check_up_to]:
            if isinstance(node2, Spatial) and node2.rank_variable in relevent_rvs:
                rv2spatial.setdefault(node2.rank_variable, []).append(node2)
            if isinstance(node2, Temporal):
                rv2temporal.setdefault(node2.rank_variable, []).append(node2)

        for shared_rv in sorted(oset(rv2spatial) & oset(rv2temporal) & relevent_rvs):
            disallowed_combinations.append(
                (
                    tuple(id(x) for x in rv2spatial[shared_rv]),
                    tuple(id(x) for x in rv2temporal[shared_rv]),
                )
            )

    if not explore_unordered_spatial_loops:
        disallowed_combinations = [x[1:] for x in disallowed_combinations]

    for combo in itertools.product(*disallowed_combinations):
        combo = oset.union(oset(), *combo)
        yield [n for n in mapping if id(n) not in combo]


def pad_with_bottom_loops(mapping: list[MappingNode], einsum: Einsum):
    rank_variables = einsum.rank_variables
    rank_var_to_count = defaultdict(lambda: 0)
    for node in mapping:
        if isinstance(node, Temporal):
            rank_var_to_count[node.rank_variable] += 1

    for rank_var in rank_variables:
        if rank_var_to_count[rank_var] < 2:
            mapping.append(Temporal(rank_variable=rank_var))


def _timeloop_style_even(mapping: list[MappingNode]):
    # Iterate through the mapping. If there are >2 TensorHolder nodes for the same
    # memory, move all below the 2nd to the same level as the 2nd.
    mapping = copy.deepcopy(mapping)
    memory2indices = defaultdict(list)
    i = 0
    while i < len(mapping):
        node = mapping[i]
        if not isinstance(mapping[i], TensorHolder):
            i += 1
            continue
        node: TensorHolder
        seen = memory2indices[node.component]
        mapping[i]._lower = False  # Lowering might re-uneven the reservationsxs

        if len(seen) <= 1:
            seen.append(i)
        else:
            mapping.insert(seen[-1] + 1, mapping.pop(i))
        i += 1
    return mapping


def assert_proper_fusion_labeling(
    mapping: list[MappingNode],
    fusable_tensors: set[TensorName],
    check_loops: bool = True,
):
    tensors = oset()
    for i, t in enumerate(mapping):
        if not isinstance(t, TensorHolder):
            continue

        new = (oset(t.tensors) - tensors) & fusable_tensors

        if new and check_loops:
            for j in range(i):
                if isinstance(mapping[j], Loop):
                    assert mapping[
                        j
                    ]._fused, f"Node {j} is not fused in {' '.join(m.compact_str() for m in mapping)}"
        assert (
            t._backing & fusable_tensors
        ) == new, f"Node {i} backing missing {new - t._backing} in {' '.join(m.compact_str() for m in mapping)}"
        tensors.update(new)
        tensors.update(t.tensors)


def iterate_mappings_no_constraints(
    spec: Spec,
    einsum_name: str,
    flattened_arch: list[arch.Leaf],
    rank_variable_bounds: dict[RankVariable, int],
    job: Job,
) -> Iterator[tuple[Mapping, SymbolTable, arch.Compute, int]]:
    first_memory = None
    for node in flattened_arch:
        if isinstance(node, arch.Memory):
            first_memory = node
            break
    if first_memory is None:
        raise ValueError("No memory found in architecture")

    einsum = spec.workload.einsums[einsum_name]
    symbol_table = {r.name: r.source for r in einsum.renames}
    fusable_tensors = job.fusable_tensors

    ranks_with_tile_pattern = oset(
        r for r, c in job.initial_delta_choices.items() if len(c) > 1
    )
    job.ranks_with_tile_pattern = ranks_with_tile_pattern

    fanouts = {}
    fanout = 1
    for node in flattened_arch:
        fanouts[node.name] = (fanout := fanout * node.get_fanout())

    for mapping, symbol_table, compute in get_tensor_choices(
        einsum_name,
        flattened_arch,
        symbol_table,
        spec,
        first_memory,
        fusable_tensors,
        fanouts,
        spec.mapper.prioritize_reuse_of_unfused_tensors,
    ):
        logging.info(
            "\tGenerated tensor choices: " + ", ".join(m.compact_str() for m in mapping)
        )
        mapping = copy.deepcopy(mapping)
        for mapping, n_orders in insert_temporal_loops(
            mapping,
            einsum,
            first_memory,
            rank_variable_bounds,
            ranks_with_tile_pattern,
            spec.workload,
            spec.mapper._can_lower_outermost_memory,
            flattened_arch,
            spec.mapper.max_fused_loops,
            fanouts,
            fusable_tensors,
            job.intermediate_tensors,
            spec.mapper._let_non_intermediate_tensors_respawn_in_backing_storage,
            spec.mapper.explore_loop_orders,
        ):
            mapping = copy.deepcopy(mapping)
            insert_spatial_loops(
                mapping, einsum, flattened_arch, job.intermediate_tensors
            )
            mapping = unpack_loops_to_rank_variables(mapping)
            if spec.mapper._timeloop_style_even:
                mapping = _timeloop_style_even(mapping)

            place_missing_temporal_loops(mapping, einsum, flattened_arch)
            label_fused_loops(mapping, fusable_tensors)
            assert_proper_fusion_labeling(mapping, fusable_tensors)
            yield mapping, symbol_table, compute, n_orders


def iterate_mappings_constraints(
    spec: Spec,
    einsum_names: list[str] | str,
    flattened_arch: list[arch.Leaf],
    rank_variable_bounds: dict[RankVariable, int],
    tensor_to_relevancy: dict[
        TensorName, dict[RankVariable, Relevant | PartiallyRelevant]
    ],
    job: Job,
) -> Iterator[tuple[Mapping, MappingConstraints, dict[str, str]]]:
    compute_name = flattened_arch[-1].name

    n_yielded = 0

    if isinstance(einsum_names, str):
        einsum_names = [einsum_names]

    for einsum_name in einsum_names:
        logging.info(
            f"Generating pmapping templates for compute {compute_name} Einsums "
            f"{einsum_name}"
        )

        for mapping, symbol_table, compute, n_orders in iterate_mappings_no_constraints(
            spec,
            einsum_name,
            flattened_arch,
            rank_variable_bounds,
            job,
        ):
            mapping, constraints = get_constraints(
                flattened_arch, mapping, symbol_table, einsum_name, tensor_to_relevancy
            )

            # This goes after the constraints because constraints may remove some loops,
            # giving us fewer that may be reordered.
            for mapping in remove_unordered_spatial_temporal_loops(
                mapping,
                flattened_arch,
                spec.workload.einsums[einsum_name],
                spec.mapper.out_of_order_hierarchy_explore_removing_spatials_for_more_temporals,
            ):
                constraints.remove_missing_targets(mapping)

                mapping.append(
                    Compute(
                        einsum=einsum_name,
                        component=compute_name,
                        component_object=compute,
                    )
                )

                # MAPPING MUST NOT BE MODIFIED AFTER constraints.set_loop_indices
                constraints.set_loop_indices(mapping)

                mapping = Mapping(nodes=[copy.copy(n) for n in mapping])
                mapping._n_loop_orders = n_orders
                yield mapping, constraints, symbol_table
                n_yielded += 1
                if n_yielded >= spec.mapper.max_pmapping_templates_per_einsum:
                    if spec.mapper._only_output_pmapping_with_index is None:
                        return
                    if (
                        isinstance(spec.mapper._only_output_pmapping_with_index, dict)
                        and einsum_name
                        not in spec.mapper._only_output_pmapping_with_index
                    ):
                        return


# =================================================================================================
# Top level
# =================================================================================================
def make_pmapping_templates(job: Job, print_progress: bool = True) -> SameEinsumJobs:
    compute_name = job.flattened_arch[-1].name

    job.tensor_to_relevancy = {
        tensor: get_rank_variable_relevancy(
            job.spec_one_einsum.workload.einsums[job.einsum_name], tensor
        )
        for tensor in job.spec_one_einsum.workload.einsums[job.einsum_name].tensor_names
    }

    mappings_constraints = iterate_mappings_constraints(
        job.spec_one_einsum,
        job.einsum_name,
        job.flattened_arch,
        job.rank_variable_bounds,
        job.tensor_to_relevancy,
        job,
    )
    if print_progress:
        mappings_constraints = tqdm(
            mappings_constraints,
            desc=f"Generating pmapping templates for compute {compute_name} Einsum {job.einsum_name}",
        )

    jobs = SameEinsumJobs()
    only_output_index = job.spec_one_einsum.mapper._only_output_pmapping_with_index
    if isinstance(only_output_index, dict):
        only_output_index = only_output_index.get(job.einsum_name, None)

    for i, (mapping, constraints, symbol_table) in enumerate(mappings_constraints):
        if only_output_index is not None and i != only_output_index:
            continue
        if isinstance(only_output_index, set) and i not in only_output_index:
            continue
        new_job = copy.copy(job)
        new_job.mapping = mapping
        new_job.mapping._template_index = i
        new_job.constraints = constraints
        new_job.job_id = uuid.uuid4()
        new_job.rank_variable_bounds = job.rank_variable_bounds
        mapping_with_reservations = quick_insert_reservation_nodes(new_job)
        new_job.compatibility = Compatibility.from_mapping(
            mapping_with_reservations,
            new_job.fusable_tensors,
            new_job.einsum,
        )
        jobs.append(new_job)

    return jobs
